import itertools

from hazelcast.config import IndexUtil, IndexType, IndexConfig
from hazelcast.future import combine_futures, ImmediateFuture
from hazelcast.invocation import Invocation
from hazelcast.protocol import PagingPredicateHolder
from hazelcast.protocol.codec import (
    map_add_entry_listener_codec,
    map_add_entry_listener_to_key_codec,
    map_add_entry_listener_with_predicate_codec,
    map_add_entry_listener_to_key_with_predicate_codec,
    map_clear_codec,
    map_contains_key_codec,
    map_contains_value_codec,
    map_delete_codec,
    map_entry_set_codec,
    map_entries_with_predicate_codec,
    map_evict_codec,
    map_evict_all_codec,
    map_flush_codec,
    map_force_unlock_codec,
    map_get_codec,
    map_get_all_codec,
    map_get_entry_view_codec,
    map_is_empty_codec,
    map_is_locked_codec,
    map_key_set_codec,
    map_key_set_with_predicate_codec,
    map_load_all_codec,
    map_load_given_keys_codec,
    map_lock_codec,
    map_put_codec,
    map_put_all_codec,
    map_put_if_absent_codec,
    map_put_transient_codec,
    map_size_codec,
    map_remove_codec,
    map_remove_if_same_codec,
    map_remove_entry_listener_codec,
    map_replace_codec,
    map_replace_if_same_codec,
    map_set_codec,
    map_try_lock_codec,
    map_try_put_codec,
    map_try_remove_codec,
    map_unlock_codec,
    map_values_codec,
    map_values_with_predicate_codec,
    map_add_interceptor_codec,
    map_aggregate_codec,
    map_aggregate_with_predicate_codec,
    map_project_codec,
    map_project_with_predicate_codec,
    map_execute_on_all_keys_codec,
    map_execute_on_key_codec,
    map_execute_on_keys_codec,
    map_execute_with_predicate_codec,
    map_add_near_cache_invalidation_listener_codec,
    map_add_index_codec,
    map_set_ttl_codec,
    map_entries_with_paging_predicate_codec,
    map_key_set_with_paging_predicate_codec,
    map_values_with_paging_predicate_codec,
    map_put_with_max_idle_codec,
    map_put_if_absent_with_max_idle_codec,
    map_put_transient_with_max_idle_codec,
    map_set_with_max_idle_codec,
)
from hazelcast.proxy.base import (
    Proxy,
    EntryEvent,
    EntryEventType,
    get_entry_listener_flags,
    MAX_SIZE,
)
from hazelcast.predicate import PagingPredicate
from hazelcast.util import (
    check_not_none,
    thread_id,
    to_millis,
    ImmutableLazyDataList,
    IterationType,
)
from hazelcast import six


class Map(Proxy):
    """Hazelcast Map client proxy to access the map on the cluster.

    Concurrent, distributed, observable and queryable map.
    This map can work both async(non-blocking) or sync(blocking).
    Blocking calls return the value of the call and block the execution until return value is calculated.
    However, async calls return ``hazelcast.future.Future`` and do not block execution.
    Result of the ``hazelcast.future.Future`` can be used whenever ready.
    A ``hazelcast.future.Future``'s result can be obtained with blocking the execution by
    calling ``future.result()``.

    Example:
        >>> my_map = client.get_map("my_map").blocking()  # sync map, all operations are blocking
        >>> print("map.put", my_map.put("key", "value"))
        >>> print("map.contains_key", my_map.contains_key("key"))
        >>> print("map.get", my_map.get("key"))
        >>> print("map.size", my_map.size())

    Example:
        >>> my_map = client.get_map("map")  # async map, all operations are non-blocking
        >>> def put_callback(f):
        >>>     print("map.put", f.result())
        >>> my_map.put("key", "async_val").add_done_callback(put_callback)
        >>>
        >>> print("map.size", my_map.size().result())
        >>>
        >>> def contains_key_callback(f):
        >>>     print("map.contains_key", f.result())
        >>> my_map.contains_key("key").add_done_callback(contains_key_callback)

    This class does not allow ``None`` to be used as a key or value.
    """

    def __init__(self, service_name, name, context):
        super(Map, self).__init__(service_name, name, context)
        self._reference_id_generator = context.lock_reference_id_generator

    def add_entry_listener(
        self,
        include_value=False,
        key=None,
        predicate=None,
        added_func=None,
        removed_func=None,
        updated_func=None,
        evicted_func=None,
        evict_all_func=None,
        clear_all_func=None,
        merged_func=None,
        expired_func=None,
        loaded_func=None,
    ):
        """Adds a continuous entry listener for this map.

        Listener will get notified for map events filtered with given parameters.

        Args:
            include_value (bool): Whether received event should include the value or not.
            key: Key for filtering the events.
            predicate (hazelcast.predicate.Predicate): Predicate for filtering the events.
            added_func (function): Function to be called when an entry is added to map.
            removed_func (function): Function to be called when an entry is removed from map.
            updated_func (function): Function to be called when an entry is updated.
            evicted_func (function): Function to be called when an entry is evicted from map.
            evict_all_func (function): Function to be called when entries are evicted from map.
            clear_all_func (function): Function to be called when entries are cleared from map.
            merged_func (function): Function to be called when WAN replicated entry is merged_func.
            expired_func (function): Function to be called when an entry's live time is expired.
            loaded_func (function): Function to be called when an entry is loaded from a map loader.

        Returns:
            hazelcast.future.Future[str]: A registration id which is used as a key to remove the listener.
        """
        flags = get_entry_listener_flags(
            ADDED=added_func,
            REMOVED=removed_func,
            UPDATED=updated_func,
            EVICTED=evicted_func,
            EXPIRED=expired_func,
            EVICT_ALL=evict_all_func,
            CLEAR_ALL=clear_all_func,
            MERGED=merged_func,
            LOADED=loaded_func,
        )

        if key and predicate:
            codec = map_add_entry_listener_to_key_with_predicate_codec
            key_data = self._to_data(key)
            predicate_data = self._to_data(predicate)
            request = codec.encode_request(
                self.name, key_data, predicate_data, include_value, flags, self._is_smart
            )
        elif key and not predicate:
            codec = map_add_entry_listener_to_key_codec
            key_data = self._to_data(key)
            request = codec.encode_request(
                self.name, key_data, include_value, flags, self._is_smart
            )
        elif not key and predicate:
            codec = map_add_entry_listener_with_predicate_codec
            predicate = self._to_data(predicate)
            request = codec.encode_request(
                self.name, predicate, include_value, flags, self._is_smart
            )
        else:
            codec = map_add_entry_listener_codec
            request = codec.encode_request(self.name, include_value, flags, self._is_smart)

        def handle_event_entry(
            key_, value, old_value, merging_value, event_type, uuid, number_of_affected_entries
        ):
            event = EntryEvent(
                self._to_object,
                key_,
                value,
                old_value,
                merging_value,
                event_type,
                uuid,
                number_of_affected_entries,
            )

            if event.event_type == EntryEventType.ADDED:
                added_func(event)
            elif event.event_type == EntryEventType.REMOVED:
                removed_func(event)
            elif event.event_type == EntryEventType.UPDATED:
                updated_func(event)
            elif event.event_type == EntryEventType.EVICTED:
                evicted_func(event)
            elif event.event_type == EntryEventType.EVICT_ALL:
                evict_all_func(event)
            elif event.event_type == EntryEventType.CLEAR_ALL:
                clear_all_func(event)
            elif event.event_type == EntryEventType.MERGED:
                merged_func(event)
            elif event.event_type == EntryEventType.EXPIRED:
                expired_func(event)
            elif event.event_type == EntryEventType.LOADED:
                loaded_func(event)

        return self._register_listener(
            request,
            lambda r: codec.decode_response(r),
            lambda reg_id: map_remove_entry_listener_codec.encode_request(self.name, reg_id),
            lambda m: codec.handle(m, handle_event_entry),
        )

    def add_index(
        self, attributes=None, index_type=IndexType.SORTED, name=None, bitmap_index_options=None
    ):
        """Adds an index to this map for the specified entries so that queries can run faster.

        Example:
            Let's say your map values are Employee objects.

                >>> class Employee(IdentifiedDataSerializable):
                >>>     active = false
                >>>     age = None
                >>>     name = None
                >>>     #other fields
                >>>
                >>>     #methods

            If you query your values mostly based on age and active fields, you should consider indexing these.

                >>> employees = client.get_map("employees")
                >>> employees.add_index(attributes=["age"]) # Sorted index for range queries
                >>> employees.add_index(attributes=["active"], index_type=IndexType.HASH)) # Hash index for equality predicates

        Index attribute should either have a getter method or be public.
        You should also make sure to add the indexes before adding
        entries to this map.

        Indexing time is executed in parallel on each partition by operation threads. The Map
        is not blocked during this operation.
        The time taken in proportional to the size of the Map and the number Members.

        Until the index finishes being created, any searches for the attribute will use a full Map scan,
        thus avoiding using a partially built index and returning incorrect results.

        Args:
            attributes (list[str]): List of indexed attributes.
            index_type (int|str): Type of the index. By default, set to ``SORTED``.
                See the :class:`hazelcast.config.IndexType` for possible values.
            name (str): Name of the index.
            bitmap_index_options (dict): Bitmap index options.

                - **unique_key:** (str): The unique key attribute is used as a
                  source of values which uniquely identify each entry being
                  inserted into an index. Defaults to ``KEY_ATTRIBUTE_NAME``.
                  See the :class:`hazelcast.config.QueryConstants` for possible
                  values.
                - **unique_key_transformation** (int|str): The transformation is
                  applied to every value extracted from the unique key attribue.
                  Defaults to ``OBJECT``. See the
                  :class:`hazelcast.config.UniqueKeyTransformation` for possible
                  values.

        Returns:
            hazelcast.future.Future[None]:
        """
        d = {
            "name": name,
            "type": index_type,
            "attributes": attributes,
            "bitmap_index_options": bitmap_index_options,
        }
        config = IndexConfig.from_dict(d)
        validated = IndexUtil.validate_and_normalize(self.name, config)
        request = map_add_index_codec.encode_request(self.name, validated)
        return self._invoke(request)

    def add_interceptor(self, interceptor):
        """Adds an interceptor for this map.

        Added interceptor will intercept operations and execute user defined methods.

        Args:
            interceptor: Interceptor for the map which includes user defined methods.

        Returns:
            hazelcast.future.Future[str]: Id of registered interceptor.
        """
        interceptor_data = self._to_data(interceptor)

        request = map_add_interceptor_codec.encode_request(self.name, interceptor_data)
        return self._invoke(request, map_add_interceptor_codec.decode_response)

    def aggregate(self, aggregator, predicate=None):
        """Applies the aggregation logic on map entries and filter the result with the
        predicate, if given.

        Args:
            aggregator (hazelcast.aggregator.Aggregator): Aggregator to aggregate the
                entries with.
            predicate (hazelcast.predicate.Predicate): Predicate to filter the entries
                with.

        Returns:
            hazelcast.future.Future: The result of the aggregation.
        """
        check_not_none(aggregator, "aggregator can't be none")
        aggregator_data = self._to_data(aggregator)

        if predicate:
            if isinstance(predicate, PagingPredicate):
                raise AssertionError("Paging predicate is not supported.")

            def handler(message):
                return self._to_object(map_aggregate_with_predicate_codec.decode_response(message))

            predicate_data = self._to_data(predicate)
            request = map_aggregate_with_predicate_codec.encode_request(
                self.name, aggregator_data, predicate_data
            )
            return self._invoke(request, handler)

        def handler(message):
            return self._to_object(map_aggregate_codec.decode_response(message))

        request = map_aggregate_codec.encode_request(self.name, aggregator_data)
        return self._invoke(request, handler)

    def clear(self):
        """Clears the map.

        The ``MAP_CLEARED`` event is fired for any registered listeners.

        Returns:
            hazelcast.future.Future[None]:
        """
        request = map_clear_codec.encode_request(self.name)
        return self._invoke(request)

    def contains_key(self, key):
        """Determines whether this map contains an entry with the key.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The specified key.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this map contains an entry for the specified key,
            ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._contains_key_internal(key_data)

    def contains_value(self, value):
        """Determines whether this map contains one or more keys for the specified value.

        Args:
            value: The specified value.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this map contains an entry for the specified value,
            ``False`` otherwise.
        """
        check_not_none(value, "value can't be None")
        value_data = self._to_data(value)

        request = map_contains_value_codec.encode_request(self.name, value_data)
        return self._invoke(request, map_contains_value_codec.decode_response)

    def delete(self, key):
        """Removes the mapping for a key from this map if it is present (optional operation).

        Unlike remove(object), this operation does not return the removed value, which avoids the serialization cost of
        the returned value. If the removed value will not be used, a delete operation is preferred over a remove
        operation for better performance.

        The map will not contain a mapping for the specified key once the call returns.

        Warning:
            This method breaks the contract of EntryListener.
            When an entry is removed by delete(), it fires an ``EntryEvent`` with a ``None`` oldValue.
            Also, a listener with predicates will have ``None`` values, so only the keys can be queried
            via predicates.

        Args:
            key: Key of the mapping to be deleted.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._delete_internal(key_data)

    def entry_set(self, predicate=None):
        """Returns a list clone of the mappings contained in this map.

        Warning:
            The list is NOT backed by the map, so changes to the map are NOT reflected in the list, and vice-versa.

        Args:
            predicate (hazelcast.predicate.Predicate): Predicate for the map to filter entries.

        Returns:
            hazelcast.future.Future[list]: The list of key-value tuples in the map.
        """
        if predicate:
            if isinstance(predicate, PagingPredicate):
                codec = map_entries_with_paging_predicate_codec

                def handler(message):
                    response = codec.decode_response(message)
                    predicate.anchor_list = response["anchor_data_list"].as_anchor_list(
                        self._to_object
                    )
                    return ImmutableLazyDataList(response["response"], self._to_object)

                predicate.iteration_type = IterationType.ENTRY
                holder = PagingPredicateHolder.of(predicate, self._to_data)
                request = codec.encode_request(self.name, holder)
            else:
                codec = map_entries_with_predicate_codec

                def handler(message):
                    return ImmutableLazyDataList(codec.decode_response(message), self._to_object)

                predicate_data = self._to_data(predicate)
                request = codec.encode_request(self.name, predicate_data)
        else:
            codec = map_entry_set_codec

            def handler(message):
                return ImmutableLazyDataList(codec.decode_response(message), self._to_object)

            request = codec.encode_request(self.name)

        return self._invoke(request, handler)

    def evict(self, key):
        """Evicts the specified key from this map.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: Key to evict.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the key is evicted, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._evict_internal(key_data)

    def evict_all(self):
        """Evicts all keys from this map except the locked ones.

        The ``EVICT_ALL`` event is fired for any registered listeners.

        Returns:
            hazelcast.future.Future[None]:
        """
        request = map_evict_all_codec.encode_request(self.name)
        return self._invoke(request)

    def execute_on_entries(self, entry_processor, predicate=None):
        """Applies the user defined EntryProcessor to all the entries in the map or entries in the map which satisfies
        the predicate if provided. Returns the results mapped by each key in the map.

        Args:
            entry_processor: A stateful serializable object which represents the EntryProcessor defined on
                server side.
                This object must have a serializable EntryProcessor counter part registered on server side with the
                actual ``com.hazelcast.map.EntryProcessor`` implementation.
            predicate (hazelcast.predicate.Predicate): Predicate for filtering the entries.

        Returns:
            hazelcast.future.Future[list]: List of map entries which includes the keys and the
            results of the entry process.
        """
        if predicate:

            def handler(message):
                return ImmutableLazyDataList(
                    map_execute_with_predicate_codec.decode_response(message), self._to_object
                )

            entry_processor_data = self._to_data(entry_processor)
            predicate_data = self._to_data(predicate)
            request = map_execute_with_predicate_codec.encode_request(
                self.name, entry_processor_data, predicate_data
            )
        else:

            def handler(message):
                return ImmutableLazyDataList(
                    map_execute_on_all_keys_codec.decode_response(message), self._to_object
                )

            entry_processor_data = self._to_data(entry_processor)
            request = map_execute_on_all_keys_codec.encode_request(self.name, entry_processor_data)

        return self._invoke(request, handler)

    def execute_on_key(self, key, entry_processor):
        """Applies the user defined EntryProcessor to the entry mapped by the key.
        Returns the object which is the result of EntryProcessor's process method.

        Args:
            key: Specified key for the entry to be processed.
            entry_processor: A stateful serializable object which represents the EntryProcessor defined on
                server side.
                This object must have a serializable EntryProcessor counter part registered on server side with the
                actual ``com.hazelcast.map.EntryProcessor`` implementation.

        Returns:
            hazelcast.future.Future[any]: Result of entry process.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._execute_on_key_internal(key_data, entry_processor)

    def execute_on_keys(self, keys, entry_processor):
        """Applies the user defined EntryProcessor to the entries mapped by the collection of keys. Returns the results
        mapped by each key in the collection.

        Args:
            keys (list): Collection of the keys for the entries to be processed.
            entry_processor: A stateful serializable object which represents the EntryProcessor defined on
                server side.
                This object must have a serializable EntryProcessor counter part registered on server side with the
                actual ``com.hazelcast.map.EntryProcessor`` implementation.

        Returns:
            hazelcast.future.Future[list]: List of map entries which includes the keys
            and the results of the entry process.
        """
        key_list = []
        for key in keys:
            check_not_none(key, "key can't be None")
            key_list.append(self._to_data(key))

        if len(keys) == 0:
            return ImmediateFuture([])

        def handler(message):
            return ImmutableLazyDataList(
                map_execute_on_keys_codec.decode_response(message), self._to_object
            )

        entry_processor_data = self._to_data(entry_processor)
        request = map_execute_on_keys_codec.encode_request(
            self.name, entry_processor_data, key_list
        )
        return self._invoke(request, handler)

    def flush(self):
        """Flushes all the local dirty entries.

        Returns:
            hazelcast.future.Future[None]:
        """
        request = map_flush_codec.encode_request(self.name)
        return self._invoke(request)

    def force_unlock(self, key):
        """Releases the lock for the specified key regardless of the lock owner.

        It always successfully unlocks the key, never blocks, and returns immediately.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the actual
            implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The key to lock.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)

        request = map_force_unlock_codec.encode_request(
            self.name, key_data, self._reference_id_generator.get_and_increment()
        )
        return self._invoke_on_key(request, key_data)

    def get(self, key):
        """Returns the value for the specified key, or ``None`` if this map does not contain this key.

        Warning:
            This method returns a clone of original value, modifying the returned value does not change the
            actual value in the map. One should put modified value back to make changes visible to all nodes.

                >>> value = my_map.get(key)
                >>> value.update_some_property()
                >>> my_map.put(key,value)

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the actual
            implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The specified key.

        Returns:
            hazelcast.future.Future[any]: The value for the specified key.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._get_internal(key_data)

    def get_all(self, keys):
        """Returns the entries for the given keys.

        Warning:
            The returned map is NOT backed by the original map, so changes to the original map are NOT reflected in the
            returned map, and vice-versa.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the actual
            implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            keys (list): Keys to get.

        Returns:
            hazelcast.future.Future[list[tuple]]: List of map entries.
        """
        check_not_none(keys, "keys can't be None")
        if not keys:
            return ImmediateFuture({})

        partition_service = self._context.partition_service
        partition_to_keys = {}

        for key in keys:
            check_not_none(key, "key can't be None")
            key_data = self._to_data(key)
            partition_id = partition_service.get_partition_id(key_data)
            try:
                partition_to_keys[partition_id][key] = key_data
            except KeyError:
                partition_to_keys[partition_id] = {key: key_data}

        return self._get_all_internal(partition_to_keys)

    def get_entry_view(self, key):
        """Returns the EntryView for the specified key.

        Warning:
            This method returns a clone of original mapping, modifying the returned value does not change the
            actual value in the map. One should put modified value back to make changes visible to all nodes.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The key of the entry.

        Returns:
            hazelcast.future.Future[hazelcast.core.SimpleEntryView]: EntryView of the specified key.
        """
        check_not_none(key, "key can't be None")

        def handler(message):
            response = map_get_entry_view_codec.decode_response(message)
            entry_view = response["response"]
            if not entry_view:
                return None

            entry_view.key = self._to_object(entry_view.key)
            entry_view.value = self._to_object(entry_view.value)
            return entry_view

        key_data = self._to_data(key)
        request = map_get_entry_view_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, handler)

    def is_empty(self):
        """Returns whether this map contains no key-value mappings or not.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this map contains no key-value mappings, ``False`` otherwise.
        """
        request = map_is_empty_codec.encode_request(self.name)
        return self._invoke(request, map_is_empty_codec.decode_response)

    def is_locked(self, key):
        """Checks the lock for the specified key.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The key that is checked for lock

        Returns:
            hazelcast.future.Future[bool]: ``True`` if lock is acquired, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)

        request = map_is_locked_codec.encode_request(self.name, key_data)
        return self._invoke_on_key(request, key_data, map_is_locked_codec.decode_response)

    def key_set(self, predicate=None):
        """Returns a List clone of the keys contained in this map or
        the keys of the entries filtered with the predicate if provided.

        Warning:
            The list is NOT backed by the map, so changes to the map are NOT reflected in the list, and vice-versa.

        Args:
            predicate (hazelcast.predicate.Predicate) Predicate to filter the entries.

        Returns:
            hazelcast.future.Future[list]: A list of the clone of the keys.
        """
        if predicate:
            if isinstance(predicate, PagingPredicate):
                codec = map_key_set_with_paging_predicate_codec

                def handler(message):
                    response = codec.decode_response(message)
                    predicate.anchor_list = response["anchor_data_list"].as_anchor_list(
                        self._to_object
                    )
                    return ImmutableLazyDataList(response["response"], self._to_object)

                predicate.iteration_type = IterationType.KEY
                holder = PagingPredicateHolder.of(predicate, self._to_data)
                request = codec.encode_request(self.name, holder)
            else:
                codec = map_key_set_with_predicate_codec

                def handler(message):
                    return ImmutableLazyDataList(codec.decode_response(message), self._to_object)

                predicate_data = self._to_data(predicate)
                request = codec.encode_request(self.name, predicate_data)
        else:
            codec = map_key_set_codec

            def handler(message):
                return ImmutableLazyDataList(codec.decode_response(message), self._to_object)

            request = codec.encode_request(self.name)

        return self._invoke(request, handler)

    def load_all(self, keys=None, replace_existing_values=True):
        """Loads all keys from the store at server side or loads the given keys if provided.

        Args:
            keys (list): Keys of the entry values to load.
            replace_existing_values (bool): Whether the existing values will be replaced or not
                with those loaded from the server side MapLoader.

        Returns:
            hazelcast.future.Future[None]:
        """
        if keys:
            key_data_list = list(map(self._to_data, keys))
            return self._load_all_internal(key_data_list, replace_existing_values)
        else:
            request = map_load_all_codec.encode_request(self.name, replace_existing_values)
            return self._invoke(request)

    def lock(self, key, lease_time=None):
        """Acquires the lock for the specified key infinitely or for the specified lease time if provided.

        If the lock is not available, the current thread becomes disabled for thread scheduling purposes and lies
        dormant until the lock has been acquired.

        You get a lock whether the value is present in the map or not. Other threads (possibly on other systems) would
        block on their invoke of lock() until the non-existent key is unlocked. If the lock holder introduces the key to
        the map, the put() operation is not blocked. If a thread not holding a lock on the non-existent key tries to
        introduce the key while a lock exists on the non-existent key, the put() operation blocks until it is unlocked.

        Scope of the lock is this map only. Acquired lock is only for the key in this map.

        Locks are re-entrant; so, if the key is locked N times, it should be unlocked N times before another thread can
        acquire it.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The key to lock.
            lease_time (int): Time in seconds to wait before releasing the lock.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)

        request = map_lock_codec.encode_request(
            self.name,
            key_data,
            thread_id(),
            to_millis(lease_time),
            self._reference_id_generator.get_and_increment(),
        )
        partition_id = self._context.partition_service.get_partition_id(key_data)
        invocation = Invocation(request, partition_id=partition_id, timeout=MAX_SIZE)
        self._invocation_service.invoke(invocation)
        return invocation.future

    def project(self, projection, predicate=None):
        """Applies the projection logic on map entries and filter the result with the
        predicate, if given.

        Args:
            projection (hazelcast.projection.Projection): Projection to project the
                entries with.
            predicate (hazelcast.predicate.Predicate): Predicate to filter the entries
                with.

        Returns:
            hazelcast.future.Future: The result of the projection.
        """
        check_not_none(projection, "Projection can't be none")
        projection_data = self._to_data(projection)

        if predicate:
            if isinstance(predicate, PagingPredicate):
                raise AssertionError("Paging predicate is not supported.")

            def handler(message):
                return ImmutableLazyDataList(
                    map_project_with_predicate_codec.decode_response(message), self._to_object
                )

            predicate_data = self._to_data(predicate)
            request = map_project_with_predicate_codec.encode_request(
                self.name, projection_data, predicate_data
            )
            return self._invoke(request, handler)

        def handler(message):
            return ImmutableLazyDataList(
                map_project_codec.decode_response(message), self._to_object
            )

        request = map_project_codec.encode_request(self.name, projection_data)
        return self._invoke(request, handler)

    def put(self, key, value, ttl=None, max_idle=None):
        """Associates the specified value with the specified key in this map.

        If the map previously contained a mapping for the key, the old value is replaced by the specified value.
        If ttl is provided, entry will expire and get evicted after the ttl.

        Warning:
            This method returns a clone of the previous value, not the original (identically equal) value previously put
            into the map.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The specified key.
            value: The value to associate with the key.
            ttl (int): Maximum time in seconds for this entry to stay in the map.
                If not provided, the value configured on the server side
                configuration will be used. Setting this to ``0`` means infinite
                time-to-live.
            max_idle (int): Maximum time in seconds for this entry to stay idle
                in the map. If not provided, the value configured on the server
                side configuration will be used. Setting this to ``0`` means
                infinite max idle time.

        Returns:
            hazelcast.future.Future[any]: Previous value associated with key
            or ``None`` if there was no mapping for key.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")
        key_data = self._to_data(key)
        value_data = self._to_data(value)
        return self._put_internal(key_data, value_data, ttl, max_idle)

    def put_all(self, map):
        """Copies all of the mappings from the specified map to this map.

        No atomicity guarantees are given. In the case of a failure, some of the key-value tuples may get written,
        while others are not.

        Args:
            map (dict): Map which includes mappings to be stored in this map.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(map, "map can't be None")
        if not map:
            return ImmediateFuture(None)

        partition_service = self._context.partition_service
        partition_map = {}

        for key, value in six.iteritems(map):
            check_not_none(key, "key can't be None")
            check_not_none(value, "value can't be None")
            entry = (self._to_data(key), self._to_data(value))
            partition_id = partition_service.get_partition_id(entry[0])
            try:
                partition_map[partition_id].append(entry)
            except KeyError:
                partition_map[partition_id] = [entry]

        futures = []
        for partition_id, entry_list in six.iteritems(partition_map):
            request = map_put_all_codec.encode_request(
                self.name, entry_list, False
            )  # TODO trigger map loader
            future = self._invoke_on_partition(request, partition_id)
            futures.append(future)

        return combine_futures(futures)

    def put_if_absent(self, key, value, ttl=None, max_idle=None):
        """Associates the specified key with the given value if it is not already associated.

        If ttl is provided, entry will expire and get evicted after the ttl.

        This is equivalent to below, except that the action is performed atomically:

            >>> if not my_map.contains_key(key):
            >>>     return my_map.put(key,value)
            >>> else:
            >>>     return my_map.get(key)

        Warning:
            This method returns a clone of the previous value, not the original (identically equal) value previously put
            into the map.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: Key of the entry.
            value: Value of the entry.
            ttl (int): Maximum time in seconds for this entry to stay in the map.
                If not provided, the value configured on the server side
                configuration will be used. Setting this to ``0`` means infinite
                time-to-live.
            max_idle (int): Maximum time in seconds for this entry to stay idle
                in the map. If not provided, the value configured on the server
                side configuration will be used. Setting this to ``0`` means
                infinite max idle time.

        Returns:
            hazelcast.future.Future[any]: Old value of the entry.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        return self._put_if_absent_internal(key_data, value_data, ttl, max_idle)

    def put_transient(self, key, value, ttl=None, max_idle=None):
        """Same as ``put``, but MapStore defined at the server side will not be called.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: Key of the entry.
            value: Value of the entry.
            ttl (int): Maximum time in seconds for this entry to stay in the map.
                If not provided, the value configured on the server side
                configuration will be used. Setting this to ``0`` means infinite
                time-to-live.
            max_idle (int): Maximum time in seconds for this entry to stay idle
                in the map. If not provided, the value configured on the server
                side configuration will be used. Setting this to ``0`` means
                infinite max idle time.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        return self._put_transient_internal(key_data, value_data, ttl, max_idle)

    def remove(self, key):
        """Removes the mapping for a key from this map if it is present.

        The map will not contain a mapping for the specified key once the call returns.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: Key of the mapping to be deleted.

        Returns:
            hazelcast.future.Future[any]: The previous value associated with key,
            or ``None`` if there was no mapping for key.
        """
        check_not_none(key, "key can't be None")
        key_data = self._to_data(key)
        return self._remove_internal(key_data)

    def remove_if_same(self, key, value):
        """Removes the entry for a key only if it is currently mapped to a given value.

        This is equivalent to below, except that the action is performed atomically:

            >>> if my_map.contains_key(key) and my_map.get(key) == value:
            >>>     my_map.remove(key)
            >>>     return True
            >>> else:
            >>>     return False

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The specified key.
            value: Remove the key if it has this value.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the value was removed, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")

        key_data = self._to_data(key)
        value_data = self._to_data(value)
        return self._remove_if_same_internal_(key_data, value_data)

    def remove_entry_listener(self, registration_id):
        """Removes the specified entry listener.

        Returns silently if there is no such listener added before.

        Args:
            registration_id (str): Id of registered listener.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if registration is removed, ``False`` otherwise.
        """
        return self._deregister_listener(registration_id)

    def replace(self, key, value):
        """Replaces the entry for a key only if it is currently mapped to some value.

        This is equivalent to below, except that the action is performed atomically:

            >>> if my_map.contains_key(key):
            >>>     return my_map.put(key,value)
            >>> else:
            >>>     return None

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Warning:
            This method returns a clone of the previous value, not the original (identically equal) value previously put
            into the map.

        Args:
            key: The specified key.
            value: The value to replace the previous value.

        Returns:
            hazelcast.future.Future[any]: Previous value associated with key,
            or ``None`` if there was no mapping for key.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")

        key_data = self._to_data(key)
        value_data = self._to_data(value)

        return self._replace_internal(key_data, value_data)

    def replace_if_same(self, key, old_value, new_value):
        """Replaces the entry for a key only if it is currently mapped to a given value.

        This is equivalent to below, except that the action is performed atomically:

            >>> if my_map.contains_key(key) and my_map.get(key) == old_value:
            >>>     my_map.put(key, new_value)
            >>>     return True
            >>> else:
            >>>     return False

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: The specified key.
            old_value: Replace the key value if it is the old value.
            new_value: The new value to replace the old value.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the value was replaced, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        check_not_none(old_value, "old_value can't be None")
        check_not_none(new_value, "new_value can't be None")

        key_data = self._to_data(key)
        old_value_data = self._to_data(old_value)
        new_value_data = self._to_data(new_value)

        return self._replace_if_same_internal(key_data, old_value_data, new_value_data)

    def set(self, key, value, ttl=None, max_idle=None):
        """Puts an entry into this map.

        Similar to the put operation except that set doesn't return the old value, which is more efficient.
        If ttl is provided, entry will expire and get evicted after the ttl.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form of the key, not the
            actual implementations of ``__hash__`` and ``__eq__`` defined in key's class.

        Args:
            key: Key of the entry.
            value: Value of the entry.
            ttl (int): Maximum time in seconds for this entry to stay in the map.
                If not provided, the value configured on the server side
                configuration will be used. Setting this to ``0`` means infinite
                time-to-live.
            max_idle (int): Maximum time in seconds for this entry to stay idle
                in the map. If not provided, the value configured on the server
                side configuration will be used. Setting this to ``0`` means
                infinite max idle time.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")
        key_data = self._to_data(key)
        value_data = self._to_data(value)
        return self._set_internal(key_data, value_data, ttl, max_idle)

    def set_ttl(self, key, ttl):
        """Updates the TTL (time to live) value of the entry specified by the given key with a new TTL value.

        New TTL value is valid starting from the time this operation is invoked,
        not since the time the entry was created. If the entry does not exist or is already expired,
        this call has no effect.

        Args:
            key: The key of the map entry.
            ttl (int): Maximum time in seconds for this entry to stay in the map.
                Setting this to ``0`` means infinite time-to-live.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(key, "key can't be None")
        check_not_none(ttl, "ttl can't be None")
        key_data = self._to_data(key)
        return self._set_ttl_internal(key_data, ttl)

    def size(self):
        """Returns the number of entries in this map.

        Returns:
            hazelcast.future.Future[int]: Number of entries in this map.
        """
        request = map_size_codec.encode_request(self.name)
        return self._invoke(request, map_size_codec.decode_response)

    def try_lock(self, key, lease_time=None, timeout=0):
        """Tries to acquire the lock for the specified key.

        When the lock is not available:

        - If the timeout is not provided, the current thread doesn't wait and returns ``False`` immediately.
        - If the timeout is provided, the current thread becomes disabled for thread scheduling purposes and lies
          dormant until one of the followings happens:

            - The lock is acquired by the current thread, or
            - The specified waiting time elapses.

        If the lease time is provided, lock will be released after this time elapses.

        Args:
            key: Key to lock in this map.
            lease_time (int): Time in seconds to wait before releasing the lock.
            timeout (int): Maximum time in seconds to wait for the lock.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the lock was acquired, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")

        key_data = self._to_data(key)
        request = map_try_lock_codec.encode_request(
            self.name,
            key_data,
            thread_id(),
            to_millis(lease_time),
            to_millis(timeout),
            self._reference_id_generator.get_and_increment(),
        )
        partition_id = self._context.partition_service.get_partition_id(key_data)
        invocation = Invocation(
            request,
            partition_id=partition_id,
            timeout=MAX_SIZE,
            response_handler=map_try_lock_codec.decode_response,
        )
        self._invocation_service.invoke(invocation)
        return invocation.future

    def try_put(self, key, value, timeout=0):
        """Tries to put the given key and value into this map and returns immediately if timeout is not provided.

        If timeout is provided, operation waits until it is completed or timeout is reached.

        Args:
            key: Key of the entry.
            value: Value of the entry.
            timeout (int): Maximum time in seconds to wait.

        Returns:
            hazelcast.future.Future[bool] ``True`` if the put is successful, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")

        key_data = self._to_data(key)
        value_data = self._to_data(value)

        return self._try_put_internal(key_data, value_data, timeout)

    def try_remove(self, key, timeout=0):
        """Tries to remove the given key from this map and returns immediately if timeout is not provided.

        If timeout is provided, operation waits until it is completed or timeout is reached.

        Args:
            key: Key of the entry to be deleted.
            timeout (int): Maximum time in seconds to wait.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the remove is successful, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")

        key_data = self._to_data(key)
        return self._try_remove_internal(key_data, timeout)

    def unlock(self, key):
        """Releases the lock for the specified key.

        It never blocks and returns immediately. If the current thread is the holder of this lock,
        then the hold count is decremented. If the hold count is zero, then the lock is released.

        Args:
            key: The key to lock.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(key, "key can't be None")

        key_data = self._to_data(key)
        request = map_unlock_codec.encode_request(
            self.name, key_data, thread_id(), self._reference_id_generator.get_and_increment()
        )
        return self._invoke_on_key(request, key_data)

    def values(self, predicate=None):
        """Returns a list clone of the values contained in this map or values of the entries which are filtered with
        the predicate if provided.

        Warning:
            The list is NOT backed by the map, so changes to the map are NOT reflected in the list, and
            vice-versa.

        Args:
            predicate (hazelcast.predicate.Predicate): Predicate to filter the entries.

        Returns:
            hazelcast.future.Future[list]: A list of clone of the values contained in this map.
        """
        if predicate:
            if isinstance(predicate, PagingPredicate):
                codec = map_values_with_paging_predicate_codec

                def handler(message):
                    response = codec.decode_response(message)
                    predicate.anchor_list = response["anchor_data_list"].as_anchor_list(
                        self._to_object
                    )
                    return ImmutableLazyDataList(response["response"], self._to_object)

                predicate.iteration_type = IterationType.VALUE
                holder = PagingPredicateHolder.of(predicate, self._to_data)
                request = codec.encode_request(self.name, holder)
            else:
                codec = map_values_with_predicate_codec

                def handler(message):
                    return ImmutableLazyDataList(codec.decode_response(message), self._to_object)

                predicate_data = self._to_data(predicate)
                request = codec.encode_request(self.name, predicate_data)
        else:
            codec = map_values_codec

            def handler(message):
                return ImmutableLazyDataList(codec.decode_response(message), self._to_object)

            request = codec.encode_request(self.name)

        return self._invoke(request, handler)

    # internals
    def _contains_key_internal(self, key_data):
        request = map_contains_key_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, map_contains_key_codec.decode_response)

    def _get_internal(self, key_data):
        def handler(message):
            return self._to_object(map_get_codec.decode_response(message))

        request = map_get_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, handler)

    def _get_all_internal(self, partition_to_keys, futures=None):
        if futures is None:
            futures = []

        def handler(message):
            return ImmutableLazyDataList(
                map_get_all_codec.decode_response(message), self._to_object
            )

        for partition_id, key_dict in six.iteritems(partition_to_keys):
            request = map_get_all_codec.encode_request(self.name, six.itervalues(key_dict))
            future = self._invoke_on_partition(request, partition_id, handler)
            futures.append(future)

        def merge(f):
            return dict(itertools.chain(*f.result()))

        return combine_futures(futures).continue_with(merge)

    def _remove_internal(self, key_data):
        def handler(message):
            return self._to_object(map_remove_codec.decode_response(message))

        request = map_remove_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, handler)

    def _remove_if_same_internal_(self, key_data, value_data):
        request = map_remove_if_same_codec.encode_request(
            self.name, key_data, value_data, thread_id()
        )
        return self._invoke_on_key(
            request, key_data, response_handler=map_remove_if_same_codec.decode_response
        )

    def _delete_internal(self, key_data):
        request = map_delete_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data)

    def _put_internal(self, key_data, value_data, ttl, max_idle):
        def handler(message):
            return self._to_object(map_put_codec.decode_response(message))

        if max_idle is not None:
            request = map_put_with_max_idle_codec.encode_request(
                self.name, key_data, value_data, thread_id(), to_millis(ttl), to_millis(max_idle)
            )
        else:
            request = map_put_codec.encode_request(
                self.name, key_data, value_data, thread_id(), to_millis(ttl)
            )
        return self._invoke_on_key(request, key_data, handler)

    def _set_internal(self, key_data, value_data, ttl, max_idle):
        if max_idle is not None:
            request = map_set_with_max_idle_codec.encode_request(
                self.name, key_data, value_data, thread_id(), to_millis(ttl), to_millis(max_idle)
            )
        else:
            request = map_set_codec.encode_request(
                self.name, key_data, value_data, thread_id(), to_millis(ttl)
            )
        return self._invoke_on_key(request, key_data)

    def _set_ttl_internal(self, key_data, ttl):
        request = map_set_ttl_codec.encode_request(self.name, key_data, to_millis(ttl))
        return self._invoke_on_key(request, key_data, map_set_ttl_codec.decode_response)

    def _try_remove_internal(self, key_data, timeout):
        request = map_try_remove_codec.encode_request(
            self.name, key_data, thread_id(), to_millis(timeout)
        )
        return self._invoke_on_key(request, key_data, map_try_remove_codec.decode_response)

    def _try_put_internal(self, key_data, value_data, timeout):
        request = map_try_put_codec.encode_request(
            self.name, key_data, value_data, thread_id(), to_millis(timeout)
        )
        return self._invoke_on_key(request, key_data, map_try_put_codec.decode_response)

    def _put_transient_internal(self, key_data, value_data, ttl, max_idle):
        if max_idle is not None:
            request = map_put_transient_with_max_idle_codec.encode_request(
                self.name, key_data, value_data, thread_id(), to_millis(ttl), to_millis(max_idle)
            )
        else:
            request = map_put_transient_codec.encode_request(
                self.name, key_data, value_data, thread_id(), to_millis(ttl)
            )
        return self._invoke_on_key(request, key_data)

    def _put_if_absent_internal(self, key_data, value_data, ttl, max_idle):
        def handler(message):
            return self._to_object(map_put_if_absent_codec.decode_response(message))

        if max_idle is not None:
            request = map_put_if_absent_with_max_idle_codec.encode_request(
                self.name, key_data, value_data, thread_id(), to_millis(ttl), to_millis(max_idle)
            )
        else:
            request = map_put_if_absent_codec.encode_request(
                self.name, key_data, value_data, thread_id(), to_millis(ttl)
            )
        return self._invoke_on_key(request, key_data, handler)

    def _replace_if_same_internal(self, key_data, old_value_data, new_value_data):
        request = map_replace_if_same_codec.encode_request(
            self.name, key_data, old_value_data, new_value_data, thread_id()
        )
        return self._invoke_on_key(request, key_data, map_replace_if_same_codec.decode_response)

    def _replace_internal(self, key_data, value_data):
        def handler(message):
            return self._to_object(map_replace_codec.decode_response(message))

        request = map_replace_codec.encode_request(self.name, key_data, value_data, thread_id())
        return self._invoke_on_key(request, key_data, handler)

    def _evict_internal(self, key_data):
        request = map_evict_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, map_evict_codec.decode_response)

    def _load_all_internal(self, key_data_list, replace_existing_values):
        request = map_load_given_keys_codec.encode_request(
            self.name, key_data_list, replace_existing_values
        )
        return self._invoke(request)

    def _execute_on_key_internal(self, key_data, entry_processor):
        def handler(message):
            return self._to_object(map_execute_on_key_codec.decode_response(message))

        entry_processor_data = self._to_data(entry_processor)
        request = map_execute_on_key_codec.encode_request(
            self.name, entry_processor_data, key_data, thread_id()
        )
        return self._invoke_on_key(request, key_data, handler)


class MapFeatNearCache(Map):
    """Map proxy implementation featuring Near Cache"""

    def __init__(self, service_name, name, context):
        super(MapFeatNearCache, self).__init__(service_name, name, context)
        self._invalidation_listener_id = None
        self._near_cache = context.near_cache_manager.get_or_create_near_cache(name)
        if self._near_cache.invalidate_on_change:
            self._add_near_cache_invalidation_listener()

    def clear(self):
        self._near_cache._clear()
        return super(MapFeatNearCache, self).clear()

    def evict_all(self):
        self._near_cache.clear()
        return super(MapFeatNearCache, self).evict_all()

    def load_all(self, keys=None, replace_existing_values=True):
        if keys is None and replace_existing_values:
            self._near_cache.clear()
        return super(MapFeatNearCache, self).load_all(keys, replace_existing_values)

    def _on_destroy(self):
        self._remove_near_cache_invalidation_listener()
        self._near_cache.clear()
        super(MapFeatNearCache, self)._on_destroy()

    def _add_near_cache_invalidation_listener(self):
        codec = map_add_near_cache_invalidation_listener_codec
        request = codec.encode_request(self.name, EntryEventType.INVALIDATION, self._is_smart)
        self._invalidation_listener_id = self._register_listener(
            request,
            lambda r: codec.decode_response(r),
            lambda reg_id: map_remove_entry_listener_codec.encode_request(self.name, reg_id),
            lambda m: codec.handle(m, self._handle_invalidation, self._handle_batch_invalidation),
        ).result()

    def _remove_near_cache_invalidation_listener(self):
        if self._invalidation_listener_id:
            self.remove_entry_listener(self._invalidation_listener_id)

    def _handle_invalidation(self, key, source_uuid, partition_uuid, sequence):
        # key is always ``Data``
        # null key means near cache has to remove all entries in it.
        # see MapAddNearCacheEntryListenerMessageTask.
        if key is None:
            self._near_cache._clear()
        else:
            self._invalidate_cache(key)

    def _handle_batch_invalidation(self, keys, source_uuids, partition_uuids, sequences):
        # key_list is always list of ``Data``
        for key_data in keys:
            self._invalidate_cache(key_data)

    def _invalidate_cache(self, key_data):
        self._near_cache._invalidate(key_data)

    def _invalidate_cache_batch(self, key_data_list):
        for key_data in key_data_list:
            self._near_cache._invalidate(key_data)

    # internals
    def _contains_key_internal(self, key_data):
        try:
            return self._near_cache[key_data]
        except KeyError:
            return super(MapFeatNearCache, self)._contains_key_internal(key_data)

    def _get_internal(self, key_data):
        try:
            value = self._near_cache[key_data]
            return ImmediateFuture(value)
        except KeyError:
            future = super(MapFeatNearCache, self)._get_internal(key_data)
            return future.continue_with(self._update_cache, key_data)

    def _update_cache(self, f, key_data):
        self._near_cache.__setitem__(key_data, f.result())
        return f.result()

    def _get_all_internal(self, partition_to_keys, futures=None):
        if futures is None:
            futures = []
        for key_dic in six.itervalues(partition_to_keys):
            for key in list(key_dic.keys()):
                try:
                    key_data = key_dic[key]
                    value = self._near_cache[key_data]
                    future = ImmediateFuture((key, value))
                    futures.append(future)
                    del key_dic[key]
                except KeyError:
                    pass
        return super(MapFeatNearCache, self)._get_all_internal(partition_to_keys, futures)

    def _try_remove_internal(self, key_data, timeout):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._try_remove_internal(key_data, timeout)

    def _try_put_internal(self, key_data, value_data, timeout):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._try_put_internal(key_data, value_data, timeout)

    def _set_internal(self, key_data, value_data, ttl, max_idle):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._set_internal(key_data, value_data, ttl, max_idle)

    def _set_ttl_internal(self, key_data, ttl):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._set_ttl_internal(key_data, ttl)

    def _replace_internal(self, key_data, value_data):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._replace_internal(key_data, value_data)

    def _replace_if_same_internal(self, key_data, old_value_data, new_value_data):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._replace_if_same_internal(
            key_data, old_value_data, new_value_data
        )

    def _remove_internal(self, key_data):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._remove_internal(key_data)

    def _remove_if_same_internal_(self, key_data, value_data):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._remove_if_same_internal_(key_data, value_data)

    def _put_transient_internal(self, key_data, value_data, ttl, max_idle):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._put_transient_internal(
            key_data, value_data, ttl, max_idle
        )

    def _put_internal(self, key_data, value_data, ttl, max_idle):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._put_internal(key_data, value_data, ttl, max_idle)

    def _put_if_absent_internal(self, key_data, value_data, ttl, max_idle):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._put_if_absent_internal(
            key_data, value_data, ttl, max_idle
        )

    def _load_all_internal(self, key_data_list, replace_existing_values):
        self._invalidate_cache_batch(key_data_list)
        return super(MapFeatNearCache, self)._load_all_internal(
            key_data_list, replace_existing_values
        )

    def _execute_on_key_internal(self, key_data, entry_processor):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._execute_on_key_internal(key_data, entry_processor)

    def _evict_internal(self, key_data):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._evict_internal(key_data)

    def _delete_internal(self, key_data):
        self._invalidate_cache(key_data)
        return super(MapFeatNearCache, self)._delete_internal(key_data)


def create_map_proxy(service_name, name, context):
    near_cache_config = context.config.near_caches.get(name, None)
    if near_cache_config is None:
        return Map(service_name, name, context)
    else:
        return MapFeatNearCache(service_name, name, context)



from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.codec.custom.raft_group_id_codec import RaftGroupIdCodec

# hex: 0x1E0100
_REQUEST_MESSAGE_TYPE = 1966336
# hex: 0x1E0101
_RESPONSE_MESSAGE_TYPE = 1966337

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request(proxy_name):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    StringCodec.encode(buf, proxy_name, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    msg.next_frame()
    return RaftGroupIdCodec.decode(msg)



import functools
import logging
import random

from hazelcast.future import Future
from hazelcast.proxy.base import Proxy
from hazelcast.cluster import VectorClock
from hazelcast.protocol.codec import (
    pn_counter_add_codec,
    pn_counter_get_codec,
    pn_counter_get_configured_replica_count_codec,
)
from hazelcast.errors import NoDataMemberInClusterError
from hazelcast.six.moves import range

_logger = logging.getLogger(__name__)


class PNCounter(Proxy):
    """PN (Positive-Negative) CRDT counter.

    The counter supports adding and subtracting values as well as
    retrieving the current counter value.
    Each replica of this counter can perform operations locally without
    coordination with the other replicas, thus increasing availability.
    The counter guarantees that whenever two nodes have received the
    same set of updates, possibly in a different order, their state is
    identical, and any conflicting updates are merged automatically.
    If no new updates are made to the shared state, all nodes that can
    communicate will eventually have the same data.

    When invoking updates from the client, the invocation is remote.
    This may lead to indeterminate state - the update may be applied but the
    response has not been received. In this case, the caller will be notified
    with a TargetDisconnectedError.

    The read and write methods provide monotonic read and RYW (read-your-write)
    guarantees. These guarantees are session guarantees which means that if
    no replica with the previously observed state is reachable, the session
    guarantees are lost and the method invocation will throw a
    ConsistencyLostError. This does not mean
    that an update is lost. All of the updates are part of some replica and
    will be eventually reflected in the state of all other replicas. This
    exception just means that you cannot observe your own writes because
    all replicas that contain your updates are currently unreachable.
    After you have received a ConsistencyLostError, you can either
    wait for a sufficiently up-to-date replica to become reachable in which
    case the session can be continued or you can reset the session by calling
    the reset() method. If you have called the reset() method,
    a new session is started with the next invocation to a CRDT replica.

    Notes:
        The CRDT state is kept entirely on non-lite (data) members. If there
        aren't any and the methods here are invoked on a lite member, they will
        fail with an NoDataMemberInClusterError.
    """

    _EMPTY_ADDRESS_LIST = []

    def __init__(self, service_name, name, context):
        super(PNCounter, self).__init__(service_name, name, context)
        self._observed_clock = VectorClock()
        self._max_replica_count = 0
        self._current_target_replica_address = None

    def get(self):
        """Returns the current value of the counter.

        Returns:
            hazelcast.future.Future[int]: The current value of the counter.

        Raises:
            NoDataMemberInClusterError: if the cluster does not contain any data members.
            ConsistencyLostError: if the session guarantees have been lost.
        """
        return self._invoke_internal(pn_counter_get_codec)

    def get_and_add(self, delta):
        """Adds the given value to the current value and returns the previous value.

        Args:
          delta (int): The value to add.

        Returns:
            hazelcast.future.Future[int]: The previous value.

        Raises:
            NoDataMemberInClusterError: if the cluster does not contain any data members.
            ConsistencyLostError: if the session guarantees have been lost.
        """

        return self._invoke_internal(pn_counter_add_codec, delta=delta, get_before_update=True)

    def add_and_get(self, delta):
        """Adds the given value to the current value and returns the updated value.

        Args:
            delta (int): The value to add.

        Returns:
            hazelcast.future.Future[int]: The updated value.

        Raises:
            NoDataMemberInClusterError: if the cluster does not contain any data members.
            ConsistencyLostError: if the session guarantees have been lost.
        """

        return self._invoke_internal(pn_counter_add_codec, delta=delta, get_before_update=False)

    def get_and_subtract(self, delta):
        """Subtracts the given value from the current value and returns the previous value.

        Args:
            delta (int): The value to subtract.

        Returns:
            hazelcast.future.Future[int]: The previous value.

        Raises:
            NoDataMemberInClusterError: if the cluster does not contain any data members.
            ConsistencyLostError: if the session guarantees have been lost.
        """

        return self._invoke_internal(pn_counter_add_codec, delta=-1 * delta, get_before_update=True)

    def subtract_and_get(self, delta):
        """Subtracts the given value from the current value and returns the updated value.

        Args:
            delta (int): The value to subtract.

        Returns:
            hazelcast.future.Future[int]: The updated value.

        Raises:
            NoDataMemberInClusterError: if the cluster does not contain any data members.
            ConsistencyLostError: if the session guarantees have been lost.
        """

        return self._invoke_internal(
            pn_counter_add_codec, delta=-1 * delta, get_before_update=False
        )

    def get_and_decrement(self):
        """Decrements the counter value by one and returns the previous value.

        Returns:
            hazelcast.future.Future[int]: The previous value.

        Raises:
            NoDataMemberInClusterError: if the cluster does not contain any data members.
            ConsistencyLostError: if the session guarantees have been lost.
        """

        return self._invoke_internal(pn_counter_add_codec, delta=-1, get_before_update=True)

    def decrement_and_get(self):
        """Decrements the counter value by one and returns the updated value.

        Returns:
            hazelcast.future.Future[int]: The updated value.

        Raises:
            NoDataMemberInClusterError: if the cluster does not contain any data members.
            ConsistencyLostError: if the session guarantees have been lost.
        """

        return self._invoke_internal(pn_counter_add_codec, delta=-1, get_before_update=False)

    def get_and_increment(self):
        """Increments the counter value by one and returns the previous value.

        Returns:
            hazelcast.future.Future[int]: The previous value.

        Raises:
            NoDataMemberInClusterError: if the cluster does not contain any data members.
            ConsistencyLostError: if the session guarantees have been lost.
        """

        return self._invoke_internal(pn_counter_add_codec, delta=1, get_before_update=True)

    def increment_and_get(self):
        """Increments the counter value by one and returns the updated value.

        Returns:
            hazelcast.future.Future[int]: The updated value.

        Raises:
            NoDataMemberInClusterError: if the cluster does not contain any data members.
            UnsupportedOperationError: if the cluster version is less than 3.10.
            ConsistencyLostError: if the session guarantees have been lost.
        """

        return self._invoke_internal(pn_counter_add_codec, delta=1, get_before_update=False)

    def reset(self):
        """Resets the observed state by this PN counter.

        This method may be used after a method invocation has thrown a ``ConsistencyLostError``
        to reset the proxy and to be able to start a new session.
        """

        self._observed_clock = VectorClock()

    def _invoke_internal(self, codec, **kwargs):
        delegated_future = Future()
        self._set_result_or_error(
            delegated_future, PNCounter._EMPTY_ADDRESS_LIST, None, codec, **kwargs
        )
        return delegated_future

    def _set_result_or_error(
        self, delegated_future, excluded_addresses, last_error, codec, **kwargs
    ):
        target = self._get_crdt_operation_target(excluded_addresses)
        if not target:
            if last_error:
                delegated_future.set_exception(last_error)
                return
            delegated_future.set_exception(
                NoDataMemberInClusterError(
                    "Cannot invoke operations on a CRDT because "
                    "the cluster does not contain any data members"
                )
            )
            return
        request = codec.encode_request(
            name=self.name,
            replica_timestamps=self._observed_clock.entry_set(),
            target_replica_uuid=target.uuid,
            **kwargs
        )

        future = self._invoke_on_target(request, target.uuid, codec.decode_response)

        checker_func = functools.partial(
            self._check_invocation_result,
            delegated_future=delegated_future,
            excluded_addresses=excluded_addresses,
            target=target,
            codec=codec,
            **kwargs
        )
        future.add_done_callback(checker_func)

    def _check_invocation_result(
        self, future, delegated_future, excluded_addresses, target, codec, **kwargs
    ):
        try:
            result = future.result()
            self._update_observed_replica_timestamp(result["replica_timestamps"])
            delegated_future.set_result(result["value"])
        except Exception as ex:
            _logger.exception(
                "Exception occurred while invoking operation on target %s, "
                "choosing different target",
                target,
            )
            if excluded_addresses == PNCounter._EMPTY_ADDRESS_LIST:
                excluded_addresses = []

            excluded_addresses.append(target)
            self._set_result_or_error(delegated_future, excluded_addresses, ex, codec, **kwargs)

    def _get_crdt_operation_target(self, excluded_addresses):
        if (
            self._current_target_replica_address
            and self._current_target_replica_address not in excluded_addresses
        ):
            return self._current_target_replica_address

        self._current_target_replica_address = self._choose_target_replica(excluded_addresses)
        return self._current_target_replica_address

    def _choose_target_replica(self, excluded_addresses):
        replica_addresses = self._get_replica_addresses(excluded_addresses)

        if len(replica_addresses) == 0:
            return None

        random_replica_index = random.randrange(0, len(replica_addresses))
        return replica_addresses[random_replica_index]

    def _get_replica_addresses(self, excluded_addresses):
        data_members = self._context.cluster_service.get_members(
            lambda member: not member.lite_member
        )
        replica_count = self._get_max_configured_replica_count()

        current_count = min(replica_count, len(data_members))
        replica_addresses = []

        for i in range(current_count):
            member_address = data_members[i]
            if member_address not in excluded_addresses:
                replica_addresses.append(member_address)

        return replica_addresses

    def _get_max_configured_replica_count(self):
        if self._max_replica_count > 0:
            return self._max_replica_count

        request = pn_counter_get_configured_replica_count_codec.encode_request(self.name)
        count = self._invoke(
            request, pn_counter_get_configured_replica_count_codec.decode_response
        ).result()
        self._max_replica_count = count
        return self._max_replica_count

    def _update_observed_replica_timestamp(self, observed_timestamps):
        observed_clock = self._to_vector_clock(observed_timestamps)
        if observed_clock.is_after(self._observed_clock):
            self._observed_clock = observed_clock

    def _to_vector_clock(self, timestamps):
        vector_clock = VectorClock()
        for replica_id, timestamp in timestamps:
            vector_clock.set_replica_timestamp(replica_id, timestamp)

        return vector_clock



import hazelcast.util as util
from hazelcast.errors import HazelcastSerializationError
from hazelcast.serialization import INT_SIZE_IN_BYTES, NULL_ARRAY_LENGTH
from hazelcast.serialization.api import PortableWriter
from hazelcast.serialization.output import EmptyObjectDataOutput
from hazelcast.serialization.portable.classdef import FieldType, ClassDefinitionBuilder
from hazelcast.six.moves import range


class DefaultPortableWriter(PortableWriter):
    def __init__(self, portable_serializer, out, class_definition):
        self._raw = False
        self._portable_serializer = portable_serializer
        self._out = out
        self._class_def = class_definition
        self._writen_fields = set()
        self._begin_pos = out.position()

        # room for final offset
        out.write_zero_bytes(4)
        out.write_int(class_definition.get_field_count())

        self._offset = out.position()
        # one additional for raw data
        field_indexes_length = (class_definition.get_field_count() + 1) * INT_SIZE_IN_BYTES
        out.write_zero_bytes(field_indexes_length)

    def write_byte(self, field_name, value):
        self._set_position(field_name, FieldType.BYTE)
        self._out.write_byte(value)

    def write_boolean(self, field_name, value):
        self._set_position(field_name, FieldType.BOOLEAN)
        self._out.write_boolean(value)

    def write_char(self, field_name, value):
        self._set_position(field_name, FieldType.CHAR)
        self._out.write_char(value)

    def write_short(self, field_name, value):
        self._set_position(field_name, FieldType.SHORT)
        self._out.write_short(value)

    def write_int(self, field_name, value):
        self._set_position(field_name, FieldType.INT)
        self._out.write_int(value)

    def write_long(self, field_name, value):
        self._set_position(field_name, FieldType.LONG)
        self._out.write_long(value)

    def write_float(self, field_name, value):
        self._set_position(field_name, FieldType.FLOAT)
        self._out.write_float(value)

    def write_double(self, field_name, value):
        self._set_position(field_name, FieldType.DOUBLE)
        self._out.write_double(value)

    def write_string(self, field_name, value):
        self._set_position(field_name, FieldType.STRING)
        self._out.write_string(value)

    def write_byte_array(self, field_name, values):
        self._set_position(field_name, FieldType.BYTE_ARRAY)
        self._out.write_byte_array(values)

    def write_boolean_array(self, field_name, values):
        self._set_position(field_name, FieldType.BOOLEAN_ARRAY)
        self._out.write_boolean_array(values)

    def write_char_array(self, field_name, values):
        self._set_position(field_name, FieldType.CHAR_ARRAY)
        self._out.write_char_array(values)

    def write_short_array(self, field_name, values):
        self._set_position(field_name, FieldType.SHORT_ARRAY)
        self._out.write_short_array(values)

    def write_int_array(self, field_name, values):
        self._set_position(field_name, FieldType.INT_ARRAY)
        self._out.write_int_array(values)

    def write_long_array(self, field_name, values):
        self._set_position(field_name, FieldType.LONG_ARRAY)
        self._out.write_long_array(values)

    def write_float_array(self, field_name, values):
        self._set_position(field_name, FieldType.FLOAT_ARRAY)
        self._out.write_float_array(values)

    def write_double_array(self, field_name, values):
        self._set_position(field_name, FieldType.DOUBLE_ARRAY)
        self._out.write_double_array(values)

    def write_string_array(self, field_name, values):
        self._set_position(field_name, FieldType.STRING_ARRAY)
        self._out.write_string_array(values)

    def write_portable(self, field_name, portable):
        fd = self._set_position(field_name, FieldType.PORTABLE)
        is_none = portable is None
        self._out.write_boolean(is_none)

        self._out.write_int(fd.factory_id)
        self._out.write_int(fd.class_id)

        if not is_none:
            _check_portable_attributes(fd, portable)
            self._portable_serializer.write_internal(self._out, portable)

    def write_portable_array(self, field_name, values):
        fd = self._set_position(field_name, FieldType.PORTABLE_ARRAY)
        _len = NULL_ARRAY_LENGTH if values is None else len(values)
        self._out.write_int(_len)
        self._out.write_int(fd.factory_id)
        self._out.write_int(fd.class_id)
        if _len > 0:
            _offset = self._out.position()
            self._out.write_zero_bytes(_len * 4)
            for i in range(0, _len):
                portable = values[i]
                _check_portable_attributes(fd, portable)
                _pos_val = self._out.position()
                self._out.write_int(_pos_val, _offset + i * INT_SIZE_IN_BYTES)
                self._portable_serializer.write_internal(self._out, portable)

    def write_null_portable(self, field_name, factory_id, class_id):
        self._set_position(field_name, FieldType.PORTABLE)
        self._out.write_boolean(True)
        self._out.write_int(factory_id)
        self._out.write_int(class_id)

    def get_raw_data_output(self):
        if not self._raw:
            _pos_val = self._out.position()
            # last index
            _index = self._class_def.get_field_count()
            self._out.write_int(_pos_val, self._offset + _index * INT_SIZE_IN_BYTES)
        self._raw = True
        return self._out

    def write_utf(self, field_name, value):
        self.write_string(field_name, value)

    def write_utf_array(self, field_name, values):
        self.write_string_array(field_name, values)

    # internal
    def _set_position(self, field_name, field_type):
        if self._raw:
            raise HazelcastSerializationError(
                "Cannot write Portable fields after get_raw_data_output() is called!"
            )
        fd = self._class_def.get_field(field_name)
        if fd is None:
            raise HazelcastSerializationError(
                "Invalid field name:'%s' for ClassDefinition(id:%s , version:%s )"
                % (field_name, self._class_def.class_id, self._class_def.version)
            )
        if field_name not in self._writen_fields:
            self._write_field_def(fd.index, field_name, field_type)
            self._writen_fields.add(field_name)
        else:
            raise HazelcastSerializationError("Field '%s' has already been written!" % field_name)
        return fd

    def _write_field_def(self, index, field_name, field_type):
        pos_val = self._out.position()
        position = self._offset + index * INT_SIZE_IN_BYTES
        self._out.write_int(pos_val, position)
        self._out.write_short(len(field_name))
        self._out.write_from(field_name.encode("utf-8"))
        self._out.write_byte(field_type)

    def end(self):
        # write final offset
        _pos_value = self._out.position()
        self._out.write_int(_pos_value, self._begin_pos)


def _check_portable_attributes(field_def, portable):
    if field_def.factory_id != portable.get_factory_id():
        raise HazelcastSerializationError(
            "Wrong Portable type! Generic portable types are not supported! "
            "Expected factory-id: %s, Actual factory-id: %s"
            % (field_def.factory_id, portable.get_factory_id())
        )
    if field_def.class_id != portable.get_class_id():
        raise HazelcastSerializationError(
            "Wrong Portable type! Generic portable types are not supported! "
            "Expected class-id: %s, Actual class-id: %s"
            % (field_def.class_id, portable.get_class_id())
        )


class ClassDefinitionWriter(PortableWriter):
    def __init__(
        self, portable_context, factory_id=None, class_id=None, version=None, class_def_builder=None
    ):
        self.portable_context = portable_context
        if class_def_builder is None:
            self._builder = ClassDefinitionBuilder(factory_id, class_id, version)
        else:
            self._builder = class_def_builder

    def write_byte(self, field_name, value):
        self._builder.add_byte_field(field_name)

    def write_boolean(self, field_name, value):
        self._builder.add_boolean_field(field_name)

    def write_char(self, field_name, value):
        self._builder.add_char_field(field_name)

    def write_short(self, field_name, value):
        self._builder.add_short_field(field_name)

    def write_int(self, field_name, value):
        self._builder.add_int_field(field_name)

    def write_long(self, field_name, value):
        self._builder.add_long_field(field_name)

    def write_float(self, field_name, value):
        self._builder.add_float_field(field_name)

    def write_double(self, field_name, value):
        self._builder.add_double_field(field_name)

    def write_string(self, field_name, value):
        self._builder.add_string_field(field_name)

    def write_byte_array(self, field_name, values):
        self._builder.add_byte_array_field(field_name)

    def write_boolean_array(self, field_name, values):
        self._builder.add_boolean_array_field(field_name)

    def write_char_array(self, field_name, values):
        self._builder.add_char_array_field(field_name)

    def write_short_array(self, field_name, values):
        self._builder.add_short_array_field(field_name)

    def write_int_array(self, field_name, values):
        self._builder.add_int_array_field(field_name)

    def write_long_array(self, field_name, values):
        self._builder.add_long_array_field(field_name)

    def write_float_array(self, field_name, values):
        self._builder.add_float_array_field(field_name)

    def write_double_array(self, field_name, values):
        self._builder.add_double_array_field(field_name)

    def write_string_array(self, field_name, values):
        self._builder.add_string_array_field(field_name)

    def write_portable(self, field_name, portable):
        if portable is None:
            raise HazelcastSerializationError(
                "Cannot write None portable without explicitly registering class definition!"
            )
        version = util.get_portable_version(portable, self.portable_context.portable_version)

        factory_id = portable.get_factory_id()
        class_id = portable.get_class_id()
        nested_builder = ClassDefinitionBuilder(factory_id, class_id, version)
        nested_class_def = self._create_nested_class_def(portable, nested_builder)
        self._builder.add_portable_field(field_name, nested_class_def)

    def write_portable_array(self, field_name, values):
        if values is None or len(values) == 0:
            raise HazelcastSerializationError(
                "Cannot write None portable without explicitly registering class definition!"
            )
        portable = values[0]
        _class_id = portable.get_class_id()
        for i in range(1, len(values)):
            if values[i].get_class_id() != _class_id:
                raise ValueError("Detected different class-ids in portable array!")
        version = util.get_portable_version(portable, self.portable_context.portable_version)
        nested_builder = ClassDefinitionBuilder(
            portable.get_factory_id(), portable.get_class_id(), version
        )
        nested_class_def = self._create_nested_class_def(portable, nested_builder)
        self._builder.add_portable_array_field(field_name, nested_class_def)

    def write_null_portable(self, field_name, factory_id, class_id):
        nested_class_def = self.portable_context.lookup_class_definition(
            factory_id, class_id, self.portable_context.portable_version
        )
        if nested_class_def is None:
            raise HazelcastSerializationError(
                "Cannot write None portable without explicitly registering class definition!"
            )
        self._builder.add_portable_field(field_name, nested_class_def)

    def get_raw_data_output(self):
        return EmptyObjectDataOutput()

    def write_utf(self, field_name, value):
        self.write_string(field_name, value)

    def write_utf_array(self, field_name, values):
        self.write_string_array(field_name, values)

    def _create_nested_class_def(self, portable, nested_builder):
        _writer = ClassDefinitionWriter(self.portable_context, class_def_builder=nested_builder)
        portable.write_portable(_writer)
        return self.portable_context.register_class_definition(nested_builder.build())

    def register_and_get(self):
        cd = self._builder.build()
        return self.portable_context.register_class_definition(cd)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import ListMultiFrameCodec
from hazelcast.protocol.builtin import DataCodec
from hazelcast.protocol.builtin import CodecUtil
from hazelcast.protocol.codec.custom.sql_query_id_codec import SqlQueryIdCodec
from hazelcast.protocol.codec.custom.sql_column_metadata_codec import SqlColumnMetadataCodec
from hazelcast.protocol.builtin import SqlPageCodec
from hazelcast.protocol.codec.custom.sql_error_codec import SqlErrorCodec

# hex: 0x210400
_REQUEST_MESSAGE_TYPE = 2163712
# hex: 0x210401
_RESPONSE_MESSAGE_TYPE = 2163713

_REQUEST_TIMEOUT_MILLIS_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_CURSOR_BUFFER_SIZE_OFFSET = _REQUEST_TIMEOUT_MILLIS_OFFSET + LONG_SIZE_IN_BYTES
_REQUEST_EXPECTED_RESULT_TYPE_OFFSET = _REQUEST_CURSOR_BUFFER_SIZE_OFFSET + INT_SIZE_IN_BYTES
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_EXPECTED_RESULT_TYPE_OFFSET + BYTE_SIZE_IN_BYTES
_RESPONSE_UPDATE_COUNT_OFFSET = RESPONSE_HEADER_SIZE


def encode_request(sql, parameters, timeout_millis, cursor_buffer_size, schema, expected_result_type, query_id):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_TIMEOUT_MILLIS_OFFSET, timeout_millis)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_CURSOR_BUFFER_SIZE_OFFSET, cursor_buffer_size)
    FixSizedTypesCodec.encode_byte(buf, _REQUEST_EXPECTED_RESULT_TYPE_OFFSET, expected_result_type)
    StringCodec.encode(buf, sql)
    ListMultiFrameCodec.encode_contains_nullable(buf, parameters, DataCodec.encode)
    CodecUtil.encode_nullable(buf, schema, StringCodec.encode)
    SqlQueryIdCodec.encode(buf, query_id, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    initial_frame = msg.next_frame()
    response = dict()
    response["update_count"] = FixSizedTypesCodec.decode_long(initial_frame.buf, _RESPONSE_UPDATE_COUNT_OFFSET)
    response["row_metadata"] = ListMultiFrameCodec.decode_nullable(msg, SqlColumnMetadataCodec.decode)
    response["row_page"] = CodecUtil.decode_nullable(msg, SqlPageCodec.decode)
    response["error"] = CodecUtil.decode_nullable(msg, SqlErrorCodec.decode)
    return response



import random

from hazelcast import six
from hazelcast.config import InMemoryFormat, EvictionPolicy
from hazelcast.util import current_time
from hazelcast.six.moves import range
from sys import getsizeof


def _lru_key_func(x):
    return x.last_access_time


def _lfu_key_func(x):
    return x.access_hit


def _random_key_func(_):
    return 0


_eviction_key_func = {
    EvictionPolicy.NONE: None,
    EvictionPolicy.LRU: _lru_key_func,
    EvictionPolicy.LFU: _lfu_key_func,
    EvictionPolicy.RANDOM: _random_key_func,
}


class DataRecord(object):
    """An expirable and evictable data object which represents a cache entry."""

    def __init__(self, key, value, create_time=None, ttl_seconds=None):
        self.key = key
        self.value = value
        self.create_time = create_time if create_time is not None else current_time()
        self.expiration_time = self.create_time + ttl_seconds if ttl_seconds is not None else None
        self.last_access_time = self.create_time
        self.access_hit = 0

    def is_expired(self, max_idle_seconds):
        """Determines whether this record is expired or not.

        Args:
            max_idle_seconds (float): The maximum idle time of record, maximum time after the last access time.

        Returns:
            bool: ``True`` is this record is not expired, ``False`` otherwise.
        """

        now = current_time()
        return (self.expiration_time is not None and self.expiration_time < now) or (
            max_idle_seconds is not None and self.last_access_time + max_idle_seconds < now
        )

    def __repr__(self):
        return (
            "DataRecord(key=%s, value=%s, create_time=%s, "
            "expiration_time=%s, last_access_time=%s, access_hit=%s)"
            % (
                self.key,
                self.value,
                self.create_time,
                self.expiration_time,
                self.last_access_time,
                self.access_hit,
            )
        )


class NearCache(dict):
    """NearCache is a local cache used by :class:`~hazelcast.proxy.map.MapFeatNearCache`."""

    def __init__(
        self,
        name,
        serialization_service,
        in_memory_format,
        time_to_live,
        max_idle,
        invalidate_on_change,
        eviction_policy,
        eviction_max_size,
        eviction_sampling_count=None,
        eviction_sampling_pool_size=None,
    ):
        super(NearCache, self).__init__()
        self.name = name
        self.serialization_service = serialization_service
        self.in_memory_format = in_memory_format
        self.time_to_live = time_to_live
        self.max_idle = max_idle
        self.invalidate_on_change = invalidate_on_change
        self.eviction_policy = eviction_policy
        self.eviction_max_size = eviction_max_size

        if eviction_sampling_count is None:  # None or zero
            self.eviction_sampling_count = max(eviction_max_size // 10, 1)
        elif 0 < eviction_sampling_count <= self.eviction_max_size:
            self.eviction_sampling_count = eviction_sampling_count
        else:
            self.eviction_sampling_count = self.eviction_max_size

        if eviction_sampling_pool_size is None:  # None or zero
            self.eviction_sampling_pool_size = max(eviction_max_size // 5, 1)
        elif 0 < eviction_sampling_pool_size <= self.eviction_max_size:
            self.eviction_sampling_pool_size = eviction_sampling_pool_size
        else:
            self.eviction_sampling_pool_size = self.eviction_max_size

        # internal
        self._key_func = _eviction_key_func[self.eviction_policy]
        self._eviction_candidates = list()
        self._evictions = 0
        self._expirations = 0
        self._hits = 0
        self._misses = 0
        self._invalidations = 0
        self._invalidation_requests = 0
        self._creation_time_in_seconds = current_time()

    def get_statistics(self):
        """Returns the statistics of the NearCache.

        Returns:
            dict: Dictionary that stores statistics related to this near cache.
        """
        stats = {
            "creation_time": self._creation_time_in_seconds,
            "evictions": self._evictions,
            "expirations": self._expirations,
            "misses": self._misses,
            "hits": self._hits,
            "invalidations": self._invalidations,
            "invalidation_requests": self._invalidation_requests,
            "owned_entry_count": self.__len__(),
            "owned_entry_memory_cost": getsizeof(self),
        }

        return stats

    def __setitem__(self, key, value):
        self._do_eviction_if_required()

        if self.in_memory_format == InMemoryFormat.BINARY:
            value = self.serialization_service.to_data(value)
        elif self.in_memory_format == InMemoryFormat.OBJECT:
            value = self.serialization_service.to_object(value)
        else:
            raise ValueError("Invalid in-memory format!!!")

        data_record = DataRecord(key, value, ttl_seconds=self.time_to_live)
        super(NearCache, self).__setitem__(key, data_record)

    def __getitem__(self, key):
        try:
            value_record = super(NearCache, self).__getitem__(key)
            if value_record.is_expired(self.max_idle):
                super(NearCache, self).__delitem__(key)
                raise KeyError
        except KeyError as ke:
            self._misses += 1
            raise ke

        if self.eviction_policy == EvictionPolicy.LRU:
            value_record.last_access_time = current_time()
        elif self.eviction_policy == EvictionPolicy.LFU:
            value_record.access_hit += 1
        self._hits += 1
        return (
            self.serialization_service.to_object(value_record.value)
            if self.in_memory_format == InMemoryFormat.BINARY
            else value_record.value
        )

    def _do_eviction_if_required(self):
        if not self._is_eviction_required():
            return
        new_eviction_samples = self._find_new_random_samples()
        new_eviction_samples_cleaned = self._scan_and_expire_collection(new_eviction_samples)
        if len(new_eviction_samples_cleaned) == 0:  # have nothing to expire
            return

        sorted_candidate_pool = sorted(new_eviction_samples_cleaned, key=self._key_func)
        min_size = min(self.eviction_sampling_pool_size, len(sorted_candidate_pool))
        # set new eviction candidate pool
        self._eviction_candidates = sorted_candidate_pool[:min_size]

        # did any item expired or do we need to evict
        if len(new_eviction_samples) == len(new_eviction_samples_cleaned):
            try:
                self.__delitem__(self._eviction_candidates[0].key)
                self._evictions += 1
                del self._eviction_candidates[0]
            except KeyError:
                # key may be evicted previously so just ignore it
                pass

    def _find_new_random_samples(self):
        records = list(self.values())  # has random order because of dict hash
        new_sample_pool = set(self._eviction_candidates)
        start = self._random_index()
        for i in range(start, start + self.eviction_sampling_count):
            index = i if i < len(records) else i - len(records)
            if records[index].is_expired(self.max_idle):
                self._clean_expired_record(records[index].key)
            elif (
                self._is_better_than_worse_entry(records[index])
                or len(new_sample_pool) < self.eviction_sampling_pool_size
            ):
                new_sample_pool.add(records[index])
        return new_sample_pool

    def _scan_and_expire_collection(self, records):
        new_records = []
        for record in records:
            if record.is_expired(self.max_idle):
                self._clean_expired_record(record.key)
            else:
                new_records.append(record)
        return new_records

    def _random_index(self):
        return random.randint(0, self.eviction_max_size - 1)

    def _is_better_than_worse_entry(self, data_record):
        return (
            len(self._eviction_candidates) == 0
            or (self._key_func(data_record) - self._key_func(self._eviction_candidates[-1])) < 0
        )

    def _is_eviction_required(self):
        return (
            self.eviction_policy != EvictionPolicy.NONE and self.eviction_max_size <= self.__len__()
        )

    def _clean_expired_record(self, key):
        try:
            self.__delitem__(key)
            self._expirations += 1
        except KeyError:
            # key may be evicted previously so just ignore it
            pass

    def _clear(self):
        size = self.__len__()
        self.clear()
        self._invalidations += size
        self._invalidation_requests += 1

    def _invalidate(self, key_data):
        try:
            self.__delitem__(key_data)
            self._invalidations += 1
        except KeyError:
            # There is nothing to invalidate
            pass
        self._invalidation_requests += 1

    def __repr__(self):
        return "NearCache(len=%s, evicted=%s)" % (self.__len__(), self._evictions)


class NearCacheManager(object):
    def __init__(self, config, serialization_service):
        self._config = config
        self._serialization_service = serialization_service
        self._caches = {}

    def get_or_create_near_cache(self, name):
        near_cache = self._caches.get(name, None)
        if not near_cache:
            near_cache_config = self._config.near_caches.get(name, None)
            if not near_cache_config:
                raise ValueError("Cannot find a near cache configuration with the name '%s'" % name)

            near_cache = NearCache(
                name,
                self._serialization_service,
                near_cache_config.in_memory_format,
                near_cache_config.time_to_live,
                near_cache_config.max_idle,
                near_cache_config.invalidate_on_change,
                near_cache_config.eviction_policy,
                near_cache_config.eviction_max_size,
                near_cache_config.eviction_sampling_count,
                near_cache_config.eviction_sampling_pool_size,
            )

            self._caches[name] = near_cache

        return near_cache

    def clear_near_caches(self):
        for cache in six.itervalues(self._caches):
            cache._clear()

    def destroy_near_cache(self, name):
        try:
            near_cache = self._caches.pop(name)
            near_cache.clear()
        except KeyError:
            pass

    def destroy_near_caches(self):
        for key in list(self._caches.keys()):
            self.destroy_near_cache(key)

    def list_near_caches(self):
        return list(self._caches.values())



"""
User API for Serialization.
"""


class ObjectDataOutput(object):
    """ObjectDataOutput provides an interface to convert any of primitive types or arrays of them to series of bytes and
    write them on a stream.
    """

    def write_from(self, buff, offset=None, length=None):
        """Writes the content of the buffer to this output stream.

        Args:
            buff (bytearray): Input buffer.
            offset (int): Offset of the buffer where copy begin.
            length (int): Length of data to be copied from the offset into stream.
        """
        raise NotImplementedError()

    def write_boolean(self, val):
        """Writes a bool value to this output stream.

        Single byte value 1 represent True, 0 represent False

        Args:
            val (bool): The bool to be written.
        """
        raise NotImplementedError()

    def write_byte(self, val):
        """Writes a byte value to this output stream.

        Args:
            val (int): The byte value to be written.
        """
        raise NotImplementedError()

    def write_short(self, val):
        """Writes a short value to this output stream.

        Args:
            val (int): The short value to be written.
        """
        raise NotImplementedError()

    def write_char(self, val):
        """Writes a char value to this output stream.

        Args:
            val (str): The char value to be written.
        """
        raise NotImplementedError()

    def write_int(self, val):
        """Writes a int value to this output stream.

        Args:
            val (int): The int value to be written.
        """
        raise NotImplementedError()

    def write_long(self, val):
        """Writes a long value to this output stream.

        Args:
            val (int): The long value to be written.
        """
        raise NotImplementedError()

    def write_float(self, val):
        """Writes a float value to this output stream.

        Args:
            val (float): The float value to be written.
        """
        raise NotImplementedError()

    def write_double(self, val):
        """Writes a double value to this output stream.

        Args:
            val (float): The double value to be written.
        """
        raise NotImplementedError()

    def write_bytes(self, val):
        """Writes a string to this output stream.

        Args:
            val (str): The string to be written.
        """
        raise NotImplementedError()

    def write_chars(self, val):
        """Writes every character of string to this output stream.

        Args:
            val (str): The string to be written.
        """
        raise NotImplementedError()

    def write_string(self, val):
        """Writes UTF-8 string to this output stream.

        Args:
            val (str): The UTF-8 string to be written.
        """
        raise NotImplementedError()

    def write_utf(self, val):
        """Writes UTF-8 string to this output stream.

        .. deprecated:: 4.1
            This method is deprecated and will be removed in the
            next major version. Use :func:`write_string` instead.

        Args:
            val (str): The UTF-8 string to be written.
        """
        raise NotImplementedError()

    def write_byte_array(self, val):
        """Writes a byte array to this output stream.

        Args:
            val (bytearray): The byte array to be written.
        """
        raise NotImplementedError()

    def write_boolean_array(self, val):
        """Writes a bool array to this output stream.

        Args:
            val (list[bool]): The bool array to be written.
        """
        raise NotImplementedError()

    def write_char_array(self, val):
        """Writes a char array to this output stream.

        Args:
            val  (list[str]): The char array to be written.
        """
        raise NotImplementedError()

    def write_int_array(self, val):
        """Writes a int array to this output stream.

        Args:
            val (list[int]): The int array to be written.
        """
        raise NotImplementedError()

    def write_long_array(self, val):
        """Writes a long array to this output stream.

        Args:
            val (list[int]): The long array to be written.
        """
        raise NotImplementedError()

    def write_double_array(self, val):
        """Writes a double array to this output stream.

        Args:
            val (list[float]): The double array to be written.
        """
        raise NotImplementedError()

    def write_float_array(self, val):
        """Writes a float array to this output stream.

        Args:
            val (list[float]): The float array to be written.
        """
        raise NotImplementedError()

    def write_short_array(self, val):
        """Writes a short array to this output stream.

        Args:
            val (list[int]): The short array to be written.
        """
        raise NotImplementedError()

    def write_string_array(self, val):
        """Writes a UTF-8 String array to this output stream.

        Args:
            val (list[str]): The UTF-8 String array to be written.
        """
        raise NotImplementedError()

    def write_utf_array(self, val):
        """Writes a UTF-8 String array to this output stream.

        .. deprecated:: 4.1
            This method is deprecated and will be removed in the
            next major version. Use :func:`write_string_array` instead.

        Args:
            val (list[str]): The UTF-8 String array to be written.
        """
        raise NotImplementedError()

    def write_object(self, val):
        """Writes an object to this output stream.

        Args:
            val: The object to be written.
        """
        raise NotImplementedError()

    def to_byte_array(self):
        """Returns a copy of internal byte array.

        Returns:
            bytearray: The copy of internal byte array
        """
        raise NotImplementedError()

    def get_byte_order(self):
        """Returns the order of bytes, as BIG_ENDIAN or LITTLE_ENDIAN.

        Returns:
            str:
        """
        raise NotImplementedError()


class ObjectDataInput(object):
    """ObjectDataInput provides an interface to read bytes from a stream and reconstruct it to any of primitive types or
    arrays of them.
    """

    def read_into(self, buff, offset=None, length=None):
        """Reads the content of the buffer into an array of bytes.

        Args:
            buff (bytearray): Input buffer.
            offset (int): Offset of the buffer where the read begin.
            length (int): Length of data to be read.

        Returns:
            bytearray: The read data.
        """
        raise NotImplementedError()

    def skip_bytes(self, count):
        """Skips over given number of bytes from input stream.

        Args:
            count (int): Number of bytes to be skipped.

        Returns:
            int: The actual number of bytes skipped.
        """
        raise NotImplementedError()

    def read_boolean(self):
        """Reads 1 byte from input stream and convert it to a bool value.

        Returns:
            bool: The bool value read.
        """
        raise NotImplementedError()

    def read_byte(self):
        """Reads 1 byte from input stream and returns it.

        Returns:
            int: The byte value read.
        """
        raise NotImplementedError()

    def read_unsigned_byte(self):
        """Reads 1 byte from input stream, zero-extends it and returns.

        Returns:
            int: The unsigned byte value read.
        """
        raise NotImplementedError()

    def read_short(self):
        """Reads 2 bytes from input stream and returns a short value.

        Returns:
            int: The short value read.
        """
        raise NotImplementedError()

    def read_unsigned_short(self):
        """Reads 2 bytes from input stream and returns an int value.

        Returns:
            int: The unsigned short value read.
        """
        raise NotImplementedError()

    def read_char(self):
        """Reads 2 bytes from the input stream and returns a str value.

        Returns:
            str: The char value read.
        """
        raise NotImplementedError()

    def read_int(self):
        """Reads 4 bytes from input stream and returns an int value.

        Returns:
            int: The int value read.
        """
        raise NotImplementedError()

    def read_long(self):
        """Reads 8 bytes from input stream and returns a long value.

        Returns:
            int: The int value read.
        """
        raise NotImplementedError()

    def read_float(self):
        """Reads 4 bytes from input stream and returns a float value.

        Returns:
            float: The float value read.
        """
        raise NotImplementedError()

    def read_double(self):
        """Reads 8 bytes from input stream and returns a double value.

        Returns:
            float: The double value read.
        """
        raise NotImplementedError()

    def read_string(self):
        """Reads a UTF-8 string from input stream and returns it.

        Returns:
            str: The UTF-8 string read.
        """
        raise NotImplementedError()

    def read_utf(self):
        """Reads a UTF-8 string from input stream and returns it.

        .. deprecated:: 4.1
            This method is deprecated and will be removed in the
            next major version. Use :func:`read_string` instead.

        Returns:
            str: The UTF-8 string read.
        """
        raise NotImplementedError()

    def read_byte_array(self):
        """Reads a byte array from input stream and returns it.

        Returns:
            bytearray: The byte array read.
        """
        raise NotImplementedError()

    def read_boolean_array(self):
        """Reads a bool array from input stream and returns it.

        Returns:
            list[bool]: The bool array read.
        """
        raise NotImplementedError()

    def read_char_array(self):
        """Reads a char array from input stream and returns it.

        Returns:
            list[str]: The char array read.
        """
        raise NotImplementedError()

    def read_int_array(self):
        """Reads a int array from input stream and returns it.

        Returns:
            list[int]: The int array read.
        """
        raise NotImplementedError()

    def read_long_array(self):
        """Reads a long array from input stream and returns it.

        Returns:
            list[int]: The long array read.
        """
        raise NotImplementedError()

    def read_double_array(self):
        """Reads a double array from input stream and returns it.

        Returns:
            list[float]: The double array read.
        """
        raise NotImplementedError()

    def read_float_array(self):
        """Reads a float array from input stream and returns it.

        Returns:
            list[float]: The float array read.
        """
        raise NotImplementedError()

    def read_short_array(self):
        """Reads a short array from input stream and returns it.

        Returns:
            list[int]: The short array read.
        """
        raise NotImplementedError()

    def read_string_array(self):
        """Reads a UTF-8 string array from input stream and returns it.

        Returns:
            list[str]: The UTF-8 string array read.
        """
        raise NotImplementedError()

    def read_utf_array(self):
        """Reads a UTF-8 string array from input stream and returns it.

        .. deprecated:: 4.1
            This method is deprecated and will be removed in the
            next major version. Use :func:`read_string_array` instead.

        Returns:
            list[str]: The UTF-8 string array read.
        """
        raise NotImplementedError()

    def read_object(self):
        """Reads a object from input stream and returns it.

        Returns:
            The object read.
        """
        raise NotImplementedError()

    def get_byte_order(self):
        """Returns the order of bytes, as BIG_ENDIAN or LITTLE_ENDIAN.

        Returns:
            str:
        """
        raise NotImplementedError()

    def position(self):
        """Returns current position in buffer.

        Returns:
            int: Current position in buffer.
        """
        raise NotImplementedError()

    def size(self):
        """Returns size of buffer.

        Returns:
            int: size of buffer.
        """
        raise NotImplementedError()


class IdentifiedDataSerializable(object):
    """IdentifiedDataSerializable is an alternative serialization method to Python pickle, which also avoids reflection
    during de-serialization.

    Each IdentifiedDataSerializable is created by a registered DataSerializableFactory.
    """

    def write_data(self, object_data_output):
        """Writes object fields to output stream.

        Args:
            object_data_output (hazelcast.serialization.api.ObjectDataOutput): The output.
        """
        raise NotImplementedError(
            "read_data must be implemented to serialize this IdentifiedDataSerializable"
        )

    def read_data(self, object_data_input):
        """Reads fields from the input stream.

        Args:
            object_data_input (hazelcast.serialization.api.ObjectDataInput): The input.
        """
        raise NotImplementedError(
            "read_data must be implemented to deserialize this IdentifiedDataSerializable"
        )

    def get_factory_id(self):
        """Returns DataSerializableFactory factory id for this class.

        Returns:
            int: The factory id.
        """
        raise NotImplementedError(
            "This method must return the factory ID for this IdentifiedDataSerializable"
        )

    def get_class_id(self):
        """Returns type identifier for this class. Id should be unique per DataSerializableFactory.

        Returns:
            int: The type id.
        """
        raise NotImplementedError(
            "This method must return the class ID for this IdentifiedDataSerializable"
        )


class Portable(object):
    """Portable provides an alternative serialization method.

    Instead of relying on reflection, each Portable is created by a registered PortableFactory.
    Portable serialization has the following advantages:

    - Support multiversion of the same object type.
    - Fetching individual fields without having to rely on reflection.
    - Querying and indexing support without de-serialization and/or reflection.
    """

    def write_portable(self, writer):
        """Serialize this portable object using given PortableWriter.

        Args:
            writer (hazelcast.serialization.api.PortableWriter): The PortableWriter.
        """
        raise NotImplementedError()

    def read_portable(self, reader):
        """Read portable fields using PortableReader.

        Args:
            reader (hazelcast.serialization.api.PortableReader): The PortableReader.
        """
        raise NotImplementedError()

    def get_factory_id(self):
        """Returns PortableFactory id for this portable class

        Returns:
            int: The factory id.
        """
        raise NotImplementedError()

    def get_class_id(self):
        """Returns class identifier for this portable class. Class id should be unique per PortableFactory.

        Returns:
            int: The class id.
        """
        raise NotImplementedError()


class StreamSerializer(object):
    """A base class for custom serialization."""

    def write(self, out, obj):
        """Writes object to ObjectDataOutput

        Args:
            out (hazelcast.serialization.api.ObjectDataOutput): Stream that object will be written to.
            obj: The object to be written.
        """
        raise NotImplementedError("write method must be implemented")

    def read(self, inp):
        """Reads object from objectDataInputStream

        Args:
            inp (hazelcast.serialization.api.ObjectDataInput): Stream that object will read from.

        Returns:
            The read object.
        """
        raise NotImplementedError("write method must be implemented")

    def get_type_id(self):
        """Returns typeId of serializer.

        Returns:
            int: The type id of the serializer.
        """
        raise NotImplementedError("get_type_id must be implemented")

    def destroy(self):
        """Called when instance is shutting down.

        It can be used to clear used resources.
        """
        raise NotImplementedError()


class PortableReader(object):
    """Provides a mean of reading portable fields from binary in form of Python primitives and arrays of these
    primitives, nested portable fields and array of portable fields.
    """

    def get_version(self):
        """Returns the global version of portable classes.

        Returns:
            int: Global version of portable classes.
        """
        raise NotImplementedError()

    def has_field(self, field_name):
        """Determine whether the field name exists in this portable class or not.

        Args:
            field_name (str): name of the field (does not support nested paths).

        Returns:
            bool: ``True`` if the field name exists in class, ``False`` otherwise.
        """
        raise NotImplementedError()

    def get_field_names(self):
        """Returns the set of field names on this portable class.

        Returns:
            set: Set of field names on this portable class.
        """
        raise NotImplementedError()

    def get_field_type(self, field_name):
        """Returns the field type of given field name.

        Args:
            field_name (str): Name of the field.

        Returns:
            hazelcast.serialization.portable.classdef.FieldType: The field type.
        """
        raise NotImplementedError()

    def get_field_class_id(self, field_name):
        """Returns the class id of given field.

        Args:
          field_name (str): Name of the field.

        Returns:
            int: class id of given field.
        """
        raise NotImplementedError()

    def read_int(self, field_name):
        """Reads a primitive int.

        Args:
            field_name (str): Name of the field.

        Returns:
            int: The int value read.
        """
        raise NotImplementedError()

    def read_long(self, field_name):
        """Reads a primitive long.

        Args:
            field_name (str): Name of the field.

        Returns:
            int: The long value read.
        """
        raise NotImplementedError()

    def read_string(self, field_name):
        """Reads a UTF-8 String.

        Args:
            field_name (str): Name of the field.

        Returns:
            str: The UTF-8 String read.
        """
        raise NotImplementedError()

    def read_utf(self, field_name):
        """Reads a UTF-8 String.

        .. deprecated:: 4.1
            This method is deprecated and will be removed in the
            next major version. Use :func:`read_string` instead.

        Args:
            field_name (str): Name of the field.

        Returns:
            str: The UTF-8 String read.
        """
        raise NotImplementedError()

    def read_boolean(self, field_name):
        """Reads a primitive bool.

        Args:
            field_name (str): Name of the field.

        Returns:
            bool: The bool value read.
        """
        raise NotImplementedError()

    def read_byte(self, field_name):
        """Reads a primitive byte.

        Args:
            field_name (str): Name of the field.

        Returns:
            int: The byte value read.
        """
        raise NotImplementedError()

    def read_char(self, field_name):
        """Reads a primitive char.

        Args:
            field_name (str): Name of the field.

        Returns:
            str: The char value read.
        """
        raise NotImplementedError()

    def read_double(self, field_name):
        """Reads a primitive double.

        Args:
            field_name (str): Name of the field.

        Returns:
            float: The double value read.
        """
        raise NotImplementedError()

    def read_float(self, field_name):
        """Reads a primitive float.

        Args:
            field_name (str): Name of the field.

        Returns:
            float: The float value read.
        """
        raise NotImplementedError()

    def read_short(self, field_name):
        """Reads a primitive short.

        Args:
            field_name (str): Name of the field.

        Returns:
            int: The short value read.
        """
        raise NotImplementedError()

    def read_portable(self, field_name):
        """Reads a portable.

        Args:
            field_name (str): Name of the field.

        Returns:
            hazelcast.serialization.api.Portable: The portable read.
        """
        raise NotImplementedError()

    def read_byte_array(self, field_name):
        """Reads a primitive byte array.

        Args:
            field_name (str): Name of the field.

        Returns:
            bytearray: The byte array read.
        """
        raise NotImplementedError()

    def read_boolean_array(self, field_name):
        """Reads a primitive bool array.

        Args:
            field_name (str): Name of the field.

        Returns:
            list[bool]): The bool array read.
        """
        raise NotImplementedError()

    def read_char_array(self, field_name):
        """Reads a primitive char array.

        Args:
            field_name (str): Name of the field.

        Returns:
            list[str]): The char array read.
        """
        raise NotImplementedError()

    def read_int_array(self, field_name):
        """Reads a primitive int array.

        Args:
            field_name (str): Name of the field.

        Returns:
            list[int]: The int array read.
        """
        raise NotImplementedError()

    def read_long_array(self, field_name):
        """Reads a primitive long array.

        Args:
            field_name (str): Name of the field.

        Returns:
            list[int]: The long array read.
        """
        raise NotImplementedError()

    def read_double_array(self, field_name):
        """Reads a primitive double array.

        Args:
            field_name (str): Name of the field.

        Returns:
            list[float]: The double array read.
        """
        raise NotImplementedError()

    def read_float_array(self, field_name):
        """Reads a primitive float array.

        Args:
            field_name (str): Name of the field.

        Returns:
            list[float]: The float array read.
        """
        raise NotImplementedError()

    def read_short_array(self, field_name):
        """Reads a primitive short array.

        Args:
            field_name (str): Name of the field.

        Returns:
            list[int]: The short array read.
        """
        raise NotImplementedError()

    def read_string_array(self, field_name):
        """Reads a UTF-8 String array.

        Args:
            field_name (str): Name of the field.

        Returns:
            str: The UTF-8 String array read.
        """
        raise NotImplementedError()

    def read_utf_array(self, field_name):
        """Reads a UTF-8 String array.

        .. deprecated:: 4.1
            This method is deprecated and will be removed in the
            next major version. Use :func:`read_string_array` instead.

        Args:
            field_name (str): Name of the field.

        Returns:
            str: The UTF-8 String array read.
        """
        raise NotImplementedError()

    def read_portable_array(self, field_name):
        """Reads a portable array.

        Args:
            field_name (str): Name of the field.

        Returns:
            list[hazelcast.serialization.api.Portable]: The portable array read.
        """
        raise NotImplementedError()

    def get_raw_data_input(self):
        """After reading portable fields, one can read remaining fields in old fashioned way
        consecutively from the end of stream. After get_raw_data_input() called, no data can be read.

        Returns:
            hazelcast.serialization.api.ObjectDataInput: The input.
        """
        raise NotImplementedError()


class PortableWriter(object):
    """Provides a mean of writing portable fields to a binary in form of Python primitives and arrays of these
    primitives, nested portable fields and array of portable fields.
    """

    def write_int(self, field_name, value):
        """Writes a primitive int.

        Args:
            field_name (str): Name of the field.
            value (int): Int value to be written.
        """
        raise NotImplementedError()

    def write_long(self, field_name, value):
        """Writes a primitive long.

        Args:
            field_name (str): Name of the field.
            value (int): Long value to be written.
        """
        raise NotImplementedError()

    def write_string(self, field_name, value):
        """Writes an UTF string.

        Args:
            field_name (str): Name of the field.
            value (str): UTF string value to be written.
        """
        raise NotImplementedError()

    def write_utf(self, field_name, value):
        """Writes an UTF string.

        .. deprecated:: 4.1
            This method is deprecated and will be removed in the
            next major version. Use :func:`write_string` instead.

        Args:
            field_name (str): Name of the field.
            value (str): UTF string value to be written.
        """
        raise NotImplementedError()

    def write_boolean(self, field_name, value):
        """Writes a primitive bool.

        Args:
            field_name (str): Name of the field.
            value (bool): Bool value to be written.
        """
        raise NotImplementedError()

    def write_byte(self, field_name, value):
        """Writes a primitive byte.

        Args:
            field_name (str): Name of the field.
            value (int): Byte value to be written.
        """
        raise NotImplementedError()

    def write_char(self, field_name, value):
        """Writes a primitive char.

        Args:
            field_name (str): Name of the field.
            value (str): Char value to be written.
        """
        raise NotImplementedError()

    def write_double(self, field_name, value):
        """Writes a primitive double.

        Args:
            field_name (str): Name of the field.
            value (float): Double value to be written.
        """
        raise NotImplementedError()

    def write_float(self, field_name, value):
        """Writes a primitive float.

        Args:
            field_name (str): Name of the field.
            value (float): Float value to be written.
        """
        raise NotImplementedError()

    def write_short(self, field_name, value):
        """Writes a primitive short.

        Args:
            field_name (str): Name of the field.
            value (int): Short value to be written.
        """
        raise NotImplementedError()

    def write_portable(self, field_name, portable):
        """Writes a Portable.

        Args:
            field_name (str): Name of the field.
            portable (hazelcast.serialization.api.Portable): Portable to be written.
        """
        raise NotImplementedError()

    def write_null_portable(self, field_name, factory_id, class_id):
        """To write a null portable value, user needs to provide class and factory ids of related class.

        Args:
            field_name (str): Name of the field.
            factory_id (int): Factory id of related portable class.
            class_id (int): Class id of related portable class.
        """
        raise NotImplementedError()

    def write_byte_array(self, field_name, values):
        """Writes a primitive byte array.

        Args:
            field_name (str): Name of the field.
            values (bytearray): Bytearray to be written.
        """
        raise NotImplementedError()

    def write_boolean_array(self, field_name, values):
        """Writes a primitive bool array.

        Args:
            field_name (str): Name of the field.
            values (list[bool]): Bool array to be written.
        """
        raise NotImplementedError()

    def write_char_array(self, field_name, values):
        """Writes a primitive char array.

        Args:
            field_name (str): Name of the field.
            values (list[str]): Char array to be written.
        """
        raise NotImplementedError()

    def write_int_array(self, field_name, values):
        """Writes a primitive int array.

        Args:
            field_name (str): Name of the field.
            values (list[int]): Int array to be written.
        """
        raise NotImplementedError()

    def write_long_array(self, field_name, values):
        """Writes a primitive long array.

        Args:
            field_name (str): Name of the field.
            values (list[int]): Long array to be written.
        """
        raise NotImplementedError()

    def write_double_array(self, field_name, values):
        """Writes a primitive double array.

        Args:
            field_name (str): Name of the field.
            values (list[float]): Double array to be written.
        """
        raise NotImplementedError()

    def write_float_array(self, field_name, values):
        """Writes a primitive float array.

        Args:
            field_name (str): Name of the field.
            values (list[float]): Float array to be written.
        """
        raise NotImplementedError()

    def write_short_array(self, field_name, values):
        """Writes a primitive short array.

        Args:
            field_name (str): Name of the field.
            values: (list[int]): Short array to be written.
        """
        raise NotImplementedError()

    def write_string_array(self, field_name, values):
        """Writes a UTF-8 String array.

        Args:
            field_name (str): Name of the field.
            values: (str): UTF-8 String array to be written.
        """
        raise NotImplementedError()

    def write_utf_array(self, field_name, values):
        """Writes a UTF-8 String array.

        .. deprecated:: 4.1
            This method is deprecated and will be removed in the
            next major version. Use :func:`write_string_array` instead.

        Args:
            field_name (str): Name of the field.
            values: (str): UTF-8 String array to be written.
        """
        raise NotImplementedError()

    def write_portable_array(self, field_name, values):
        """Writes a portable array.

        Args:
            field_name (str): Name of the field.
            values (list[hazelcast.serialization.api.Portable]): Portable array to be written.
        """
        raise NotImplementedError()

    def get_raw_data_output(self):
        """After writing portable fields, one can write remaining fields in old fashioned way
        consecutively at the end of stream. After get_raw_data_output() called, no data can be written.

        Returns:
            hazelcast.serialization.api.ObjectDataOutput: The output.
        """
        raise NotImplementedError()



from hazelcast.future import ImmediateFuture, Future
from hazelcast.protocol.codec import (
    ringbuffer_add_all_codec,
    ringbuffer_add_codec,
    ringbuffer_capacity_codec,
    ringbuffer_head_sequence_codec,
    ringbuffer_read_many_codec,
    ringbuffer_read_one_codec,
    ringbuffer_remaining_capacity_codec,
    ringbuffer_size_codec,
    ringbuffer_tail_sequence_codec,
)
from hazelcast.proxy.base import PartitionSpecificProxy
from hazelcast.util import (
    check_not_negative,
    check_not_none,
    check_not_empty,
    check_true,
    ImmutableLazyDataList,
)

OVERFLOW_POLICY_OVERWRITE = 0
"""
Configuration property for DEFAULT overflow policy. When an item is tried to be added on full Ringbuffer, oldest item in
the Ringbuffer is overwritten and item is added.
"""

OVERFLOW_POLICY_FAIL = 1
"""
Configuration property for overflow policy. When an item is tried to be added on full Ringbuffer, the call fails and
item is not added.

The reason that FAIL exist is to give the opportunity to obey the ttl. If blocking behavior is required, this can be
implemented using retrying in combination with an exponential backoff.

    >>> sleepMS = 100;
    >>> while true:
    >>>     result = ringbuffer.add(item, -1)
    >>>     if result != -1:
    >>>         break
    >>>     sleep(sleepMS / 1000)
    >>>     sleepMS *= 2
"""

MAX_BATCH_SIZE = 1000
"""
The maximum number of items to be added to RingBuffer or read from RingBuffer at a time.
"""


class Ringbuffer(PartitionSpecificProxy):
    """A Ringbuffer is an append-only data-structure where the content is
    stored in a ring like structure.

    A ringbuffer has a capacity so it won't grow beyond that capacity and
    endanger the stability of the system. If that capacity is exceeded, than
    the oldest item in the ringbuffer is overwritten. The ringbuffer has two
    always incrementing sequences:

    - :func:`tail_sequence`: This is the side where the youngest item is found.
      So the tail is the side of the ringbuffer where items are added to.
    - :func:`head_sequence`: This is the side where the oldest items are found.
      So the head is the side where items gets discarded.

    The items in the ringbuffer can be found by a sequence that is in between
    (inclusive) the head and tail sequence.

    If data is read from a ringbuffer with a sequence that is smaller than the
    head sequence, it means that the data is not available anymore and a
    :class:`hazelcast.errors.StaleSequenceError` is thrown.

    A Ringbuffer currently is a replicated, but not partitioned data structure.
    So all data is stored in a single partition, similarly to the
    :class:`hazelcast.proxy.queue.Queue` implementation.

    A Ringbuffer can be used in a way similar to the Queue, but one of the key
    differences is that a :func:`hazelcast.proxy.queue.Queue.take` is destructive,
    meaning that only 1 thread is able to take an item. A :func:`read_one` is not
    destructive, so you can have multiple threads reading the same item multiple
    times.
    """

    def __init__(self, service_name, name, context):
        super(Ringbuffer, self).__init__(service_name, name, context)
        self._capacity = None

    def capacity(self):
        """Returns the capacity of this Ringbuffer.

        Returns:
            hazelcast.future.Future[int]: The capacity of Ringbuffer.
        """
        if not self._capacity:

            def handler(message):
                self._capacity = ringbuffer_capacity_codec.decode_response(message)
                return self._capacity

            request = ringbuffer_capacity_codec.encode_request(self.name)
            return self._invoke(request, handler)

        return ImmediateFuture(self._capacity)

    def size(self):
        """Returns number of items in the Ringbuffer.

        Returns:
            hazelcast.future.Future[int]: The size of Ringbuffer.
        """
        request = ringbuffer_size_codec.encode_request(self.name)
        return self._invoke(request, ringbuffer_size_codec.decode_response)

    def tail_sequence(self):
        """Returns the sequence of the tail.

        The tail is the side of the Ringbuffer where the items are added to.
        The initial value of the tail is ``-1``.

        Returns:
            hazelcast.future.Future[int]: The sequence of the tail.
        """
        request = ringbuffer_tail_sequence_codec.encode_request(self.name)
        return self._invoke(request, ringbuffer_tail_sequence_codec.decode_response)

    def head_sequence(self):
        """Returns the sequence of the head.

        The head is the side of the Ringbuffer where the oldest items in the
        Ringbuffer are found. If the Ringbuffer is empty, the head will be one
        more than the tail. The initial value of the head is ``0`` (``1`` more
        than tail).

        Returns:
            hazelcast.future.Future[int]: The sequence of the head.
        """
        request = ringbuffer_head_sequence_codec.encode_request(self.name)
        return self._invoke(request, ringbuffer_head_sequence_codec.decode_response)

    def remaining_capacity(self):
        """Returns the remaining capacity of the Ringbuffer.

        Returns:
            hazelcast.future.Future[int]: The remaining capacity of Ringbuffer.
        """
        request = ringbuffer_remaining_capacity_codec.encode_request(self.name)
        return self._invoke(request, ringbuffer_remaining_capacity_codec.decode_response)

    def add(self, item, overflow_policy=OVERFLOW_POLICY_OVERWRITE):
        """Adds the specified item to the tail of the Ringbuffer.

        If there is no space in the Ringbuffer, the action is determined by
        ``overflow_policy`` as :const:`OVERFLOW_POLICY_OVERWRITE` or
        :const:`OVERFLOW_POLICY_FAIL`.

        Args:
            item: The specified item to be added.
            overflow_policy (int): the OverflowPolicy to be used when there is
                no space.

        Returns:
            hazelcast.future.Future[int]: The sequenceId of the added item, or
            ``-1`` if the add failed.
        """
        item_data = self._to_data(item)
        request = ringbuffer_add_codec.encode_request(self.name, overflow_policy, item_data)
        return self._invoke(request, ringbuffer_add_codec.decode_response)

    def add_all(self, items, overflow_policy=OVERFLOW_POLICY_OVERWRITE):
        """Adds all of the item in the specified collection to the tail of the
        Ringbuffer.

        This is likely to outperform multiple calls to :func:`add` due
        to better io utilization and a reduced number of executed operations.
        The items are added in the order of the Iterator of the collection.

        If there is no space in the Ringbuffer, the action is determined by
        ``overflow_policy`` as :const:`OVERFLOW_POLICY_OVERWRITE` or
        :const:`OVERFLOW_POLICY_FAIL`.

        Args:
            items (list): The specified collection which contains the items
                to be added.
            overflow_policy (int): The OverflowPolicy to be used when there
                is no space.

        Returns:
            hazelcast.future.Future[int]: The sequenceId of the last written item,
            or ``-1`` of the last write is failed.
        """
        check_not_empty(items, "items can't be empty")
        if len(items) > MAX_BATCH_SIZE:
            raise AssertionError("Batch size can't be greater than %d" % MAX_BATCH_SIZE)

        item_data_list = []
        for item in items:
            check_not_none(item, "item can't be None")
            item_data_list.append(self._to_data(item))

        request = ringbuffer_add_all_codec.encode_request(
            self.name, item_data_list, overflow_policy
        )
        return self._invoke(request, ringbuffer_add_all_codec.decode_response)

    def read_one(self, sequence):
        """Reads one item from the Ringbuffer.

        If the sequence is one beyond the current tail, this call blocks until
        an item  is added. Currently it isn't possible to control how long
        this call is going to block.

        Args:
            sequence (int): The sequence of the item to read.

        Returns:
            The read item.
        """
        check_not_negative(sequence, "sequence can't be smaller than 0")

        def handler(message):
            return self._to_object(ringbuffer_read_one_codec.decode_response(message))

        request = ringbuffer_read_one_codec.encode_request(self.name, sequence)
        return self._invoke(request, handler)

    def read_many(self, start_sequence, min_count, max_count, filter=None):
        """Reads a batch of items from the Ringbuffer.

        If the number of available items after the first read item is smaller
        than the ``max_count``, these items are returned. So it could be the
        number of items read is smaller than the ``max_count``. If there are
        less items available than ``min_count``, then this call blocks.

        Warnings:
            These blocking calls consume server memory and if there are many
            calls, it can be possible to see leaking memory or
            ``OutOfMemoryError`` s on the server.

        Reading a batch of items is likely to perform better because less
        overhead is involved.

        A filter can be provided to only select items that need to be read. If
        the filter is ``None``, all items are read. If the filter is not
        ``None``, only items where the filter function returns true are
        returned. Using  filters is a good way to prevent getting items that
        are of no value to the receiver. This reduces the amount of IO and the
        number of operations being executed, and can result in a significant
        performance improvement. Note that, filtering logic must be defined
        on the server-side.

        If the ``start_sequence`` is smaller than the smallest sequence still
        available in the Ringbuffer (:func:`head_sequence`), then the smallest
        available sequence will be used as the start sequence and the
        minimum/maximum number of items will be attempted to be read from there
        on.

        If the ``start_sequence`` is bigger than the last available sequence
        in the Ringbuffer (:func:`tail_sequence`), then the last available
        sequence plus one will be used as the start sequence and the call will
        block until further items become available and it can read at least the
        minimum number of items.

        Args:
            start_sequence (int): The start sequence of the first item to read.
            min_count (int): The minimum number of items to read.
            max_count (int): The maximum number of items to read.
            filter: Filter to select returned elements.

        Returns:
            hazelcast.future.Future[ReadResult]: The list of read items.
        """
        check_not_negative(start_sequence, "sequence can't be smaller than 0")
        check_not_negative(min_count, "min count can't be smaller than 0")
        check_true(max_count >= min_count, "max count should be greater or equal to min count")
        check_true(
            max_count < MAX_BATCH_SIZE, "max count can't be greater than %d" % MAX_BATCH_SIZE
        )

        request = ringbuffer_read_many_codec.encode_request(
            self.name, start_sequence, min_count, max_count, self._to_data(filter)
        )

        def handler(message):
            response = ringbuffer_read_many_codec.decode_response(message)
            read_count = response["read_count"]
            next_seq = response["next_seq"]
            items = response["items"]
            item_seqs = response["item_seqs"]

            return ReadResult(read_count, next_seq, items, item_seqs, self._to_object)

        def continuation(future):
            # Since the first call to capacity
            # is cached on the client-side, doing
            # a capacity check each time should not
            # be a problem
            capacity = future.result()

            check_true(
                max_count <= capacity,
                "max count: %d should be smaller or equal to capacity: %d" % (max_count, capacity),
            )

            return self._invoke(request, handler)

        return self.capacity().continue_with(continuation)


class ReadResult(ImmutableLazyDataList):
    """Defines the result of a :func:`Ringbuffer.read_many` operation."""

    SEQUENCE_UNAVAILABLE = -1
    """Value returned from methods returning a sequence number when the
    information is not available (e.g. because of rolling upgrade and some
    members not returning the sequence).
    """

    def __init__(self, read_count, next_seq, items, item_seqs, to_object):
        super(ReadResult, self).__init__(items, to_object)
        self._read_count = read_count
        self._next_seq = next_seq
        self._item_seqs = item_seqs

    @property
    def read_count(self):
        """int: The number of items that have been read before filtering.

        If no filter is set, then the :attr:`read_count` will be equal to
        :attr:`size`.

        But if a filter is applied, it could be that items are read, but are
        filtered out. So, if you are trying to make another read based on
        this, then you should increment the sequence by :attr:`read_count` and
        not by :attr:`size`.

        Otherwise you will be re-reading the same filtered messages.
        """
        return self._read_count

    @property
    def size(self):
        """int: The result set size.

        See Also:
            :attr:`read_count`
        """
        return len(self._list_data)

    @property
    def next_sequence_to_read_from(self):
        """int: The sequence of the item following the last read item.

        This sequence can then be used to read items following the ones
        returned by this result set.

        Usually this sequence is equal to the sequence used to retrieve this
        result set incremented by the :attr:`read_count`. In cases when the
        reader tolerates lost items, this is not the case.

        For instance, if the reader requests an item with a stale sequence (one
        which has already been overwritten), the read will jump to the oldest
        sequence and read from there.

        Similarly, if the reader requests an item in the future (e.g. because
        the partition was lost and the reader was unaware of this), the read
        method will jump back to the newest available sequence.

        Because of these jumps and only in the case when the reader is loss
        tolerant, the next sequence must be retrieved using this method.
        A return value of :const:`SEQUENCE_UNAVAILABLE` means that the
        information is not available.
        """
        return self._next_seq

    def get_sequence(self, index):
        """Return the sequence number for the item at the given index.

        Args:
            index (int): The index.

        Returns:
            int: The sequence number for the ringbuffer item.
        """
        return self._item_seqs[index]
